"""Initialize the tech_environment_service subpackage of pade_python package"""

# allow absolute import from the root folder
# whatever its name is.
import sys  # don't remove required for error handling
import os

# Import from sibling directory ..\tech_environment_service
OS_NAME = os.name

sys.path.append("..")

if OS_NAME.lower() == "nt":
    print("windows")
    sys.path.append(os.path.dirname(os.path.abspath(__file__ + "\\..")))
    sys.path.append(os.path.dirname(os.path.abspath(__file__ + "\\..\\..")))
    sys.path.append(os.path.dirname(os.path.abspath(__file__ + "\\..\\..\\..")))
else:
    print("non windows")
    sys.path.append(os.path.dirname(os.path.abspath(__file__ + "/..")))
    sys.path.append(os.path.dirname(os.path.abspath(__file__ + "/../..")))
    sys.path.append(os.path.dirname(os.path.abspath(__file__ + "/../../..")))


import pade_python.tech_environment_service.dataset_core
import pade_python.tech_environment_service.dataset_crud
import pade_python.tech_environment_service.environment_core
import pade_python.tech_environment_service.environment_file
import pade_python.tech_environment_service.environment_spark
import pade_python.tech_environment_service.environment_logging
import pade_python.tech_environment_service.security_core
import pade_python.tech_environment_service.repo_core
import pade_python.tech_environment_service.job_core



""" Module for spark and os environment for tech_environment_service with minimal dependencies. """

import sys  # don't remove required for error handling
import os

import io

# library management
from importlib import util  # library management

# error handling
from subprocess import check_output, Popen, PIPE, CalledProcessError
import subprocess

# http
from urllib.parse import urlparse
import requests

# azcopy and adls
from azure.identity import DefaultAzureCredential, ManagedIdentityCredential, ClientSecretCredential
from azure.storage.filedatalake import DataLakeServiceClient, DataLakeDirectoryClient
from azure.keyvault.secrets import SecretClient

# spark
from pyspark.sql import (SparkSession, DataFrame)
from pyspark.sql.functions import (coalesce, col, concat_ws, lit, udf, regexp_replace)
from pyspark.sql.types import (IntegerType, LongType, StringType, StructField, StructType)

#  data
pyspark_pandas_loader = util.find_spec("pyspark.pandas")
pyspark_pandas_found = pyspark_pandas_loader is not None

if pyspark_pandas_found:
    # import pyspark.pandas  as pd
    # bug - pyspark version will not read local files in the repo
    import pyspark.pandas as pd
else:
    import pandas as pd


class EnvironmentSpark:
    """ EnvironmentSpark class with minimal dependencies for the developer service.
    - This class is used to configure the Spark environment.
    """

    @staticmethod
    def configure_spark_local(
        spark_home_path: str = "",
        pyspark_python_path: str = "",
        java_home_path: str = "",
    ) -> str:
        """Configures spark locally

        Args:
            spark_home_path (str, optional): path for spark home. Defaults to "".
            pyspark_python_path (str, optional): path for pyspark python. Defaults to "".
            java_home_path (str, optional): path for java home. Defaults to "".

        Returns:
            str: spark_home_path
        """

        os.environ["SPARK_HOME"] = spark_home_path
        os.environ["JAVA_HOME"] = java_home_path

        # Add the PySpark directories to the Python path:
        sys.path.insert(1, os.path.join(java_home_path, "/bin"))
        sys.path.insert(1, os.path.join(spark_home_path, "python"))
        sys.path.insert(1, os.path.join(spark_home_path, "python", "pyspark"))
        sys.path.insert(1, os.path.join(spark_home_path, "python", "build"))

        # If PySpark isn't specified, use currently running Python binary:
        pyspark_python_path = pyspark_python_path or sys.executable
        os.environ["PYSPARK_PYTHON"] = pyspark_python_path

        # Find Spark Locally
        # location = findspark.find()
        # findspark.init(location, edit_rc=True)
        # Start a SparkContext

        return spark_home_path

    @staticmethod
    def configure_spark_path_server(spark: SparkSession) -> str:
        """Configures spark path on server

        Args:
            spark (SparkSession): Spark Session

        Returns:
            str: root_path
        """

        spark_home_path_current = os.environ.get("SPARK_HOME", None)
        print("spark_home_path_current:{spark_home_path_current}")
        print(spark_home_path_current)
        spark.conf.set("spark.databricks.userInfoFunctions.enabled", "True")
        spark.conf.set("spark.databricks.io.cache.enabled", "True")

        df_user_name = spark.sql("select current_user()")
        row_user_name = df_user_name.first()
        if row_user_name is not None:
            username = row_user_name["current_user"]
        else:
            username = "dataframe is empty error"
        repo = "pade"
        sys_path = str(sys.path)
        root_path = os.path.abspath(f"/Workspace/Repos/{username}/{repo}/pade_python")
        if root_path not in sys_path:
            sys.path.append(root_path)

        return root_path

    @staticmethod
    def get_spark_home_path_local(spark_home_path: str) -> str:
        """Retrieve spark home path when running locally

        Args:
            spark_home_path (str): default spark home path

        Returns:
            str: spark home path
        """

        if not spark_home_path:
            for path in [
                "/usr/local/opt/apache-spark/libexec",  # OS X Homebrew
                "/usr/lib/spark/"  # AWS Amazon EMR
                # Any other common places to look?
            ]:
                if os.path.exists(path):
                    spark_home_path = path
                    break

        if not spark_home_path:
            spark_home_path = "SPARK_HOME_NOT_FOUND"

        return spark_home_path



""" Module for rep_core for it_admin_service that handles repository and cluster functions with minimal dependencies. """

import os
import sys  # don't remove required for error handling

import traceback  # don't remove required for error handling
from importlib import util  # library management

from calendar import c
import json
from html.parser import HTMLParser  # web scraping html
from string import Formatter
import base64
import csv
import requests
from io import BytesIO

from datetime import date, datetime


# http
from urllib.parse import urlparse

# certs
import certifi

class RepoCore:

    @staticmethod
    def describe_cluster_library_state(library_object, cluster_id, status):
        """Converts cluster library status response to a verbose message"""

        lib_title = library_object.title()
        lib_obj = library_object
        ni_msg = f"{lib_title} library is not installed on cluster {cluster_id}."
        inst_msg = f"{lib_obj} library is already installed on cluster {cluster_id}."
        pend_msg = "Pending installation of {} library on cluster {} . . ."
        resolve_msg = f"Retrieving metadata for the installation of {lib_obj} library on cluster {cluster_id} . . ."
        install_msg = f"Installing {lib_obj} library on cluster {cluster_id} . . ."
        failed_msg = f"{lib_title} library failed to install on cluster {cluster_id}."
        uninst_msg = f"{lib_title} library installed on cluster {cluster_id} has been marked for removal upon restart."

        result_map = {
            "NOT_INSTALLED": ni_msg,
            "INSTALLED": inst_msg,
            "PENDING": pend_msg,
            "RESOLVING": resolve_msg,
            "INSTALLING": install_msg,
            "FAILED": failed_msg,
            "UNINSTALL_ON_RESTART": uninst_msg
        }

        return result_map[status.upper()]

    @staticmethod
    def get_cicd_destination_path(config_folder: str) -> str:
        """Create file path for cicd storage event trigger

        Args:
            config (dict): configuration dictionary

        Returns:
            str: cicd destination path

        """

        cicd_action_folder = config_folder.replace('config', 'cicd')

        current_date_time = datetime.now().strftime("%Y_%m_%d_%I_%M_%S_%p")

        cicd_action_folder = cicd_action_folder.replace('abfss', 'https')
        url = urlparse(cicd_action_folder)
        container = url.netloc.split("@")[0]
        base_address = url.netloc.split("@")[1]
        path = url.path
        destination_path = f"https://{base_address}/{container}{path}pull_request_{current_date_time}.json"

        return destination_path

    @staticmethod
    def call_databricks_install_api(config, end_point, body, method="GET", verbose=True):
        """Execute HTTP request against Databricks REST API 2.0"""

        token = config['access_token']
        domain = config["pade_databricks_instance_id"]
        base_url = f"https://{domain}/api/"
        cluster_id = config['pade_databricks_cluster']

        bearer = "Bearer " + token
        headers = {"Authorization": bearer, "Content-Type": "application/json"}

        method = method.upper()
        url = base_url + end_point

        if method == "GET":
            response_install = requests.get(url, headers=headers, json=body)
        elif method == "POST":
            response_install = requests.post(url, headers=headers, json=body)

        data = None

        try:
            data = json.loads(response_install.text)
            response_install_text = json.dumps(response_install.json())
            print("- response : success  -")
            response_install_text_message = "Received Cluster API Response : "
            response_install_text_message += f"{response_install_text} when posting to : {url}  "
            response_install_text_message += f"to install library: {body}"
            response_install_text_message += f"to {response_install}"

            if response_install.status_code == 200:
                statuses = response_install.json()
                if "library_statuses" in statuses:
                    for status in statuses["library_statuses"]:
                        if status["library"] == "TODO":
                            if verbose is True:
                                msg = cls.describe_cluster_library_state("library_source", cluster_id, status["status"])
                                response_install_text_message += msg
                            else:
                                response_install_text_message += status["status"]
                        else:
                            response_install_text_message += str(status)

        except Exception as exception_check:
            html_filter = HTMLFilter()
            html_filter.feed(response_install.text)
            response_install_text_message = html_filter.text
            print(f"- response : error - {str(exception_check)}")
            print("Error SAVE-PYTHON-RESPONSE converting response")
            print(f"text:{response_install} to json")

        if data is None:
            response_install_text_message = "Error Install Library"

        return response_install_text_message

    @classmethod
    def get_cluster_library_status(cls, config, library_source, properties, verbose=True):
        """Gets the current library status """

        library_source = library_source.lower()

        # Validate input
        error_message_1 = "Error: Invalid library source specified. Valid sources are: jar, egg, whl, pypi, maven, cran"
        assert library_source in ("jar", "egg", "whl", "pypi", "maven", "cran"), error_message_1
        assert properties is not None, "Error: Empty properties provided"

        # Get the cluster ID from the Spark environment
        cluster_id = config['pade_databricks_cluster']

        # Set default result to not installed
        result = cls.describe_cluster_library_state(library_source, cluster_id, "NOT_INSTALLED")

        # Execute REST API request to get the library statuses
        endpoint = f"2.0/libraries/cluster-status?cluster_id={cluster_id}"
        result = cls.call_databricks_install_api(config, endpoint, None, "GET", verbose)

        return result

    @classmethod
    def install_cluster_library(cls, config, library_source, content_data):
        """
        Installs a cluster library given correct source and properties are provided
        For examples see https://docs.databricks.com/api/latest/libraries.html#install
        """

        library_file = content_data
        library_source = library_source.lower()

        # Validate input
        err_msg = "Error: Invalid library source specified. Valid sources are: jar, egg, whl, pypi, maven, cran"
        assert library_source in ("jar", "egg", "whl", "pypi", "maven", "cran"), err_msg
        assert library_file is not None, "Error: Empty library_file provided"

        # Get the cluster ID from the Spark environment
        cluster_id = config['pade_databricks_cluster']

        status = cls.get_cluster_library_status(config, library_source, library_file, False).upper()
        if status != "INSTALLED":
            # Create the HTTP request body based on the cluster ID, library source and properties
            libraries = f'"{library_source}": "{library_file}"'
            json_string = '{"cluster_id": "' + cluster_id + '", "libraries":[{' + libraries + '}]}'
            print(f"json_string:{json_string}")
            body = json.loads(json_string)
            # Execute REST API request to install the library
            result = cls.call_databricks_install_api(config, "2.0/libraries/install", body, "POST")
            if result == '- response : success  -':
                print("Installation started . . .")
            else:
                print(result)
        else:
            print(cls.describe_cluster_library_state(library_source, cluster_id, status))

        return status

    @staticmethod
    def write_issues(r, csvout):
        "output a list of issues to csv"
        if not r.status_code == 200:
            raise Exception(r.status_code)
        for issue in r.json():
            Tag=[]
            labels = issue['labels']
            for label in labels:
                Tag.append(label['name'])

            csvout.writerow([issue['number'], issue['title'].encode('utf-8'),Tag,issue['state'] ,issue['created_at'], issue['closed_at']])

    @classmethod
    def import_issues_to_csv_from_github(cls, config: dict) -> str:
        """Uses basic authentication (Github username + password) to retrieve Issues
        from a repository that username has access to and writes the issues to a csv file.
        Supports Github API v3.

        Args:
            config (dict): Environment configuration dictionary

        Returns:
            str: Import status message
        """

        GITHUB_USER = 'USER NAME'
        GITHUB_PASSWORD = 'PASSWORD'
        REPO = 'REPO NAME'  # format is username/repo
        ISSUES_FOR_REPO_URL = 'https://api.github.com/repos/%s/issues' % REPO
        AUTH = (GITHUB_USER, GITHUB_PASSWORD)
        arg = "?state=all"

        response = requests.get(ISSUES_FOR_REPO_URL+arg)
        file_csv = '%s-issues.csv' % (REPO.replace('/', '-'))
        file_csv_o = open(file_csv, 'wb')
        out_csv = csv.writer(file_csv_o)
        out_csv.writerow(('id', 'Title', 'Tag', 'State', 'Open Date', 'Close Date'))

        cls.write_issues(r, out_csv)

        # more pages? examine the 'link' header returned
        if 'link' in response.headers:
            pages = dict(
                [(rel[6:-1], url[url.index('<')+1:-1]) for url, rel in
                    [link.split(';') for link in
                        response.headers['link'].split(',')]])

            while 'last' in pages and 'next' in pages:
                response = requests.get(pages['next'], auth=AUTH)
                cls.write_issues(response, out_csv)
                if pages['next'] == pages['last']:
                    break
                pages = dict(
                [(rel[6:-1], url[url.index('<')+1:-1]) for url, rel in
                    [link.split(';') for link in
                        response.headers['link'].split(',')]])

        file_csv_o.close()

        return "Success"

    @staticmethod
    def import_file(config, content_data, content_type, file_path) -> bool:
        """Imports file from abfss to repo library

        Args:
            config (_type_): _description_
            content_data (_type_): _description_
            content_type (_type_): _description_
            file_path (_type_): _description_

        Returns:
            bool: _description_
        """

        environment = config["environment"]
        project_id = config["project_id"]
        project_id_root = config["project_id_root"]
        file_name = os.path.basename(file_path)
        databricks_instance_id = config["databricks_instance_id"]
        token = config['access_token']
        bearer = "Bearer " + token
        headers = {"Authorization": bearer, "Content-Type": "application/json"}
        headers_redacted = str(headers).replace(bearer, "[bearer REDACTED]")
        api_version = "/api/2.0"
        databricks_instance_id = config["databricks_instance_id"]
        api_command = "/workspace/import"
        url = f"https://{databricks_instance_id}{api_version}{api_command}"

        file_path = file_path.replace("/Workspace", "")

        print(f"content_type:{content_type}")
        print(f"url:{url}")

        if content_type == "bytes":
            headers_import = {"Authorization": bearer} # , "Content-Type": "multipart/form-data"
            headers_redacted = str(headers_import).replace(
                bearer, "[bearer REDACTED]"
            )
            try:
                content_data.decode("UTF-8")
            except (UnicodeDecodeError, AttributeError):
                content_data = bytes(content_data, 'utf-8')
                pass

            files = {"upload_file": content_data}

            multipart_form_data = {
                'path': f"{file_path}"
            }

            print(f"multipart_form_data:{str(multipart_form_data)}")
            print(f"headers:{headers_redacted}")

            # binary
            # https://dbc-a1b2345c-d6e7.cloud.databricks.com/api/2.0/workspace/import \
            # --header 'Content-Type: multipart/form-data' \
            # --form path=/Users/me@example.com/MyFolder/MyNotebook \
            # --form content=@myCode.py.zip

            response_binary = PADEObject()
            response_binary.text = 'Empty. Unable to retrieve post response'
            # data=multipart_form_data,
            try:
                response_binary = requests.post(url=url, files=files, data=multipart_form_data, headers=headers_import)
                print(f"post : success : {url} ")
                response_binary_text = json.loads(response_binary.text)
                response_binary_text = json.dumps(response_binary_text.json())
                print(f"parse : success : {url}")
                response_binary_text_message = "Received Cluster API Response : "
                response_binary_text_message += f"{response_binary_text} when posting to : {url}  "
                response_binary_text_message += f"to import file: {file_path}"
            except Exception as exception_check:
                html_filter = HTMLFilter()
                html_filter.feed(response_binary.text)
                response_install_text_message = html_filter.text
                print(f"url : error - {str(exception_check)}")
                print("Error IMPORT-FILE-RESPONSE")
                print(f"response error code:{str(response_binary)}")
                print(f"response error message:{response_install_text_message}")
        elif content_type == "text":

            # text
            # curl -n -X POST https://<databricks-instance>/api/2.0/workspace/import
            # -F path="/Users/user@example.com/new-notebook" -F format=SOURCE -F language=SCALA -F overwrite=true -F content=@notebook.scala

            headers_import = {"Authorization": bearer, "Accept": "application/json"}
            headers_redacted = str(headers_import).replace(
                bearer, "[bearer REDACTED]"
            )

            print(f"headers:{headers_redacted}")
            print(f"json:{str(content_data[ 0 : 100 ])}...")

            response_json = requests.post(
                url=url, json=content_data, headers=headers_import
            )

        return True

    @classmethod
    def pull_repository_latest(
        cls,
        config: dict,
        token: str,
        base_path: str,
        repository_name: str,
        branch_name: str
    ) -> str:
        """Pulls the lastest repository branch for the given repo

        Args:
            config (dict): global config dictionary
            token (str): security token
            base_path (str): reository base path location to pull
            repository_name (str): repository name to pull
            branch_name (str): repository branch name to pull

        Returns:
            str: result message from pull request
        """

        databricks_instance_id = config["databricks_instance_id"]
        json_text = {"path": base_path}
        headers = {"Authentication": f"Bearer {token}"}
        api_url = f"https://{databricks_instance_id}"
        url = f"{api_url}/api/2.0/workspace/list"
        verify = certifi.where()

        print(f"------- Fetch {base_path}  -------")
        print(f"url:{str(url)}")
        headers_redacted = str(headers).replace(token, "[bearer REDACTED]")
        print(f"headers:{headers_redacted}")

        response = requests.get(url=url, headers=headers, json=json_text, verify=verify)
        data = None

        try:
            data = response.json()
            response_text_fetch = f"Suceess: Received list_repos with length : {len(str(data))}"
            response_text_fetch = response_text_fetch + f" when posting to : {url}"
            print(f"{response_text_fetch}")
            print(f"listed files for : {base_path}")
            print(str(data))
            lst = data["objects"]
            repos = list(filter(lambda itm: str(Path(itm['path']).stem).upper() == repository_name.upper(), lst))

            if repos[0] is None:
                repo_data = "Error Repo Not found"
            else:
                repo_object = repos[0]
                repo_id = repo_object['object_id']
                url = f"{api_url}/api/2.0/repos/{repo_id}"
                print(f"repo_id:{repo_id} branch_name:{branch_name}")
                repo_data = requests.patch(url=url, headers=headers, verify=verify, json={'branch': branch_name}).json()
        except Exception as exception_object:
            filter_object = HTMLFilter()
            filter_object.feed(response.text)
            response_text = filter_object.text
            repo_data = f"Response : error - {exception_object}: {response_text}"

        print(repo_data)

        return repo_data

class PADEObject(object):
    pass

class HTMLFilter(HTMLParser):
    text = ""

    def handle_data(self, data):
        self.text += data



""" Module with a variety of utility function for Spark data frames. """

from collections import Counter

# libraries
from importlib import util

# spark /data
from pyspark.sql import DataFrame
from pyspark.sql.functions import (
    lit,
    udf,
    expr,
    coalesce
)

from pyspark.sql.types import StringType,StructField


# util
import hashlib

import uuid
uuid_udf = udf(lambda: str(uuid.uuid4()), StringType())

pyspark_pandas_loader = util.find_spec("pyspark.pandas")
pyspark_pandas_found = pyspark_pandas_loader is not None

if pyspark_pandas_found:
    import pyspark.pandas as pd
    # bug - pyspark version will not read local files in the repo
    # import pandas as pd
else:
    import pandas as pd


def encrypt_value(pii_col):
    """Encypts a value using Databricks encryption library.

    Args:
        pii_col (_type_): Column to encrypt.

    Returns:
        _type_: Encrypted value
    """
    sha_value = hashlib.sha1(pii_col.encode()).hexdigest()
    return sha_value

encrypt_value_udf = udf(encrypt_value, StringType())


class DataSetCore:
    """Core class for Spark Datasets

    """


    @classmethod
    def add_row_id_to_dataframe(cls, sorted_df: DataFrame, row_id_keys: str,
                                yyyy_param: str, mm_param: str, dd_param: str) -> DataFrame:
        """Adds row_id column to the dataframe, the row_id a required unique identifier used
        to perform incremntal updates.
        - Replaces {yyyy}, {mm}, or {dd} with the current year, month, or day in row id template
        - Create row_id based on template
        - If row_id_key is empty, then uses uuid to create row_id

        Args:
            sorted_df (DataFrame): Dataframe to add column
            row_id_keys (str): Comma seperated list of keys to use to generate row_id
            yyyy_param (str): Year parameter to use to generate row_id
            mm_param (str): Month parameter to use to generate row_id
            dd_param (str): Day parameter to use to generate row_id

        Returns:
            DataFrame: Dataframe to return with add row_id
        """

        if row_id_keys is None:
            row_id_keys = ""
        row_id_keys = row_id_keys.strip()
        row_id_keys_list = row_id_keys.split(",")
        if len(row_id_keys_list) > 0 and len(row_id_keys) > 0:
            sql_expr = row_id_keys
            sql_expr = sql_expr.replace("{yyyy}", yyyy_param)
            sql_expr = sql_expr.replace("{mm}", mm_param)
            sql_expr = sql_expr.replace("{dd}", dd_param)
        else:
            sql_expr = "uuid()"
        sql_expr = "concat_ws('-'," + sql_expr + ")"
        print(f"attempting to update deltalake: sql_expr: {sql_expr}")

        assert " " not in "".join(sorted_df.columns)
        results_df = sorted_df.withColumn("row_id", expr(sql_expr))

        return results_df

    @staticmethod
    def encrypt_value(pii_col):
        """Encypts value to store in databricks column

        Args:
            pii_col (any): Value to encrypt

        Returns:
            any: Encrpyted value
        """
        sha_value = hashlib.sha1(pii_col.encode()).hexdigest()
        return sha_value

    @staticmethod
    def table_exists(spark, dataset_name, database_name):
        """Verifies if a dataset exists in databricks

        Args:
            spark (_type_): spark object
            dataset_name (_type_): dataset name to check
            database_name (_type_): database name to check

        Returns:
            _type_:  True if dataset exists
        """

        datasets_list_df = spark.sql(f"SHOW TABLES from {database_name}")
        datasets_list_df = datasets_list_df.filter(
            datasets_list_df.tableName == f"{dataset_name}"
        )

        if datasets_list_df.first() is not None:
            print(f"Dataset {dataset_name} found")
            return True
        else:
            print(f"Dataset {dataset_name} not found")
            return False

    @staticmethod
    def scrub_object_name(original_object_name: str) -> str:
        """Scrubs characters in object to rename

        Args:
            original_object_name (str): original column name

        Returns:
            str: new object name
        """

        if original_object_name is None:
            original_object_name = "object_name_is_missing"

        c_renamed = original_object_name
        c_renamed = c_renamed.replace("†", "_")
        c_renamed = c_renamed.replace(",", "_")
        c_renamed = c_renamed.replace("*", "_")
        c_renamed = c_renamed.replace(" ", "_")
        c_renamed = c_renamed.replace("\r", "_")
        c_renamed = c_renamed.replace("\n", "_")
        c_renamed = c_renamed.replace(";", "")
        c_renamed = c_renamed.replace(".", "")
        c_renamed = c_renamed.replace("}", "")
        c_renamed = c_renamed.replace("{", "")
        c_renamed = c_renamed.replace("(", "")
        c_renamed = c_renamed.replace(")", "")
        c_renamed = c_renamed.replace("?", "")
        c_renamed = c_renamed.replace("-", "")
        c_renamed = c_renamed.replace("/", "")
        c_renamed = c_renamed.replace("//", "")
        c_renamed = c_renamed.replace("=", "_")
        c_renamed = c_renamed.replace("&", "w")
        c_renamed = c_renamed.lower()
        c_renamed = c_renamed.strip()

        return c_renamed

    @staticmethod
    def rename_column_names_as_unique(original_list):
        """Make all the items unique by adding a suffix (1, 2, etc).

        `seq` is mutable sequence of strings.
        `suffs` is an optional alternative suffix iterable.
        """

        new_list = []
        for i, original_col in enumerate(original_list):
            if original_col is None:
                original_col = "Column"
                resulted_counter = Counter(
                    original_list
                )  # {'foo': 2, 'bar': 1, None: 2}
                totalcount = resulted_counter[None]  # 2
                count = original_list[:i].count(None)
            elif original_col == "":
                original_col = "Column"
                resulted_counter = Counter(
                    original_list
                )  # {'foo': 2, 'bar': 1, None: 2}
                totalcount = resulted_counter[""]  # 2
                count = original_list[:i].count("")
            else:
                totalcount = original_list.count(original_col)
                count = original_list[:i].count(original_col)

            if totalcount > 1:
                new_name = original_col + str(count + 1)
            else:
                new_name = original_col

            new_list.append(new_name)
        return new_list

    @staticmethod
    def database_object_exists(spark, database_name, dataset_name=None):
        """Verifies if a dataset exists in databricks

        Args:
            spark (_type_): spark object
            dataset_name (_type_): dataset name to check
            database_name (_type_): database name to check

        Returns:
            _type_:  True if dataset exists
        """

        datasets_list_df = spark.sql(f"SHOW TABLES from {database_name}")

        if dataset_name is not None:
            datasets_list_df = datasets_list_df.filter(
                datasets_list_df.tableName == f"{dataset_name}"
            )

        dataset_name = str(dataset_name)

        if datasets_list_df.first() is not None:
            print(f"Database Object {dataset_name} found")
            return True
        else:
            print(f"Dataset Object {dataset_name} not found")
            return False

    @classmethod
    def encrpyt_pii_columns(cls, pii_columns: str, is_using_standard_column_names: str,
                            sorted_df: DataFrame) -> DataFrame:
        """ Encrypts the columns that are marked as PII

        Args:
            pii_columns (str): Comma delimited list of PII columns
            is_using_standard_column_names (str): Either None or "force_lowercase"
            sorted_df (DataFrame): Dataframe to be encrypted

        Returns:
            DataFrame: Encrypted dataframe
        """
        print(f"pii_columns:{pii_columns}")
        if pii_columns is not None:
            pii_columns_list = pii_columns.split(",")
            for col_orig in pii_columns_list:
                if is_using_standard_column_names == "force_lowercase":
                    col_orig = col_orig.lower()
                    col_orig = col_orig.replace("'", "")
                    col_orig = col_orig.replace('"', "")
                sorted_df = sorted_df.withColumn(
                    col_orig, coalesce(col_orig, lit("null"))
                ).withColumn(col_orig, encrypt_value_udf(col_orig))

        return sorted_df



""" Module for to convert differe file formats to Spark data frames. """

# core
import sys  # don't remove required for error handling
import os
import pyreadstat as prs

from urllib.parse import urlparse
from io import BytesIO

# types
import json
from collections import defaultdict

# libraries
from importlib import util

# adls and azure security
from azure.storage.filedatalake import DataLakeServiceClient
from azure.identity import ClientSecretCredential

# file excel
import openpyxl
import openpyxl.utils.cell

# PADE
import pade_python.tech_environment_service.environment_file as pade_env_file
import pade_python.tech_environment_service.dataset_core as pade_ds_core
import pade_python.tech_environment_service.security_core as pade_sec_core
import pade_python.tech_environment_service.environment_core as pade_env_core


# data
import numpy as np

# spark / data
import uuid
from pyspark.sql import SparkSession, DataFrame
import pyspark.sql.utils
import pyspark.sql.functions as f
from pyspark.sql.functions import (
    col,
    lit,
    concat_ws,
    to_date,
    coalesce,
    trim,
    lpad,
    udf,
    length,
    when,
    expr,
    to_json,
    explode,
)

from pyspark.sql.types import StructType, IntegerType, StructField, LongType, ArrayType, StringType,DoubleType,DecimalType
from pyspark.sql import Row
uuid_udf = udf(lambda: str(uuid.uuid4()), StringType())
from collections import Counter

import pandas as pd_legacy
pyspark_pandas_loader = util.find_spec("pyspark.pandas")
pyspark_pandas_found = pyspark_pandas_loader is not None

if pyspark_pandas_found:
    import pyspark.pandas as pd
    # bug - pyspark version will not read local files
    # import pandas as pd
else:
    import pandas as pd

class DataSetConvert:
    """
    DataSetConvert Class to convert a dataset from one format to another.
    - Intended primarily to convert source file formats into Spark DataFrames.
    """

    @staticmethod
    def get_json_by_key(obj_json, target: str) -> dict:
        """Get json section based on the supplied path/target string

        Args:
            obj_json (any): Json object to search and extract section from
            target (str): Target/path to search for in json object

        Returns:
            any: Returns json section based on the supplied path/target string
        """

        for key in obj_json.keys():
            print(str(key))

        if target in obj_json:
            return obj_json[target]

        for key in obj_json.keys():
            key_scrubbed = key.strip().lower()
            target_scrubbed = target.strip().lower()
            print(
                f"looking for target_scrubbed:{target_scrubbed} in key_scrubbed:{key_scrubbed} "
            )
            if key_scrubbed == target_scrubbed:
                return obj_json[key]
        return {"message": "not_found", "key": target}

    @classmethod
    def convert_csv_tsv_usv_to_dataframe(cls, spark: SparkSession,
                                         ingress_file_path: str,
                                         first_row_is_header: bool,
                                         delimiter: str,
                                         encoding: str) -> DataFrame:
        """Loads csv or usv file to dataframe
        - Infers schema from file, postal codes will not be converted
        to strings
        - Date data types are not correctly inferred, TimeStamps are inferred
        - Column names are scrubbed to remove special characters and spaces
        - Original column names are stored in the metadata column

        Args:
            spark (SparkSession): SparkSession
            ingress_file_path (str): ADLS2 path to csv or usv file
            first_row_is_header (bool): true of false indicatory if file has
            header row
            delimiter (str): column delimiter
            encoding (str): encoding type

        Returns:
            DataFrame: Converted Spark dataframe
        """

        print("attempting to load dataframe for csv/tsv/usv")
        print(f"ingress_file_path:{ingress_file_path}")

        unsorted_df = (
            spark.read.format("csv")
            .option("header", first_row_is_header)
            .option("sep", delimiter)
            .option("encoding", encoding)
            .option("inferSchema", True)
            .option("treatEmptyValuesAsNulls", "true")
            .option("forceLowercaseNames", True)
            .option("multiline", "true")
            .option("inferLong", True)
            .option("inferDecimal", True)
            .option("inferInteger", True)
            .load(
                ingress_file_path,
                forceLowercaseNames=True,
                inferLong=True,
                inferDecimal=True,
                inferInteger=True,
            )
        )

        obj_ds_core = pade_ds_core.DataSetCore

        for col_orig in unsorted_df.columns:
            c_renamed = obj_ds_core.scrub_object_name(col_orig)
            unsorted_df = unsorted_df.withColumnRenamed(col_orig, c_renamed)
            unsorted_df = unsorted_df.withMetadata(c_renamed, {"ingress_column_name": col_orig, 'comment': col_orig})
            print(f"{c_renamed} :metadata: {unsorted_df.schema[c_renamed].metadata}")

        return unsorted_df

    @classmethod
    def convert_json_to_dataframe(cls, spark: SparkSession, ingress_file_path: str, encoding: str,
                                  source_json_path: str, source_dataset_name: str, dataset_name: str) -> DataFrame:
        """Loads json file to dataframe
        - Infers schema from file, postal codes will not be converted to strings
        - Date data types are not correctly inferred, TimeStamps are inferred
        - Column names are created based on the supplied configuration
        - Original column names are stored in the metadata column
        - The system currently supports nested and flat json.
        - The structure of the json is inferred based on the number of rows/columns found at the specified path

        Args:
            spark (SparkSession): SparkSession
            ingress_file_path (str): ADLS2 path to json file
            encoding (str): Encoding type
            source_json_path (str): Json path to section of json file to be loaded
            source_dataset_name (str): Name of the source dataset - name in json
            dataset_name (str): Name of the dataset/table to create

        Returns:
            DataFrame: Converted Spark dataframe
        """

        print("attempting to load dataframe for json")
        print(f"ingress_file_path:{ingress_file_path}")
        json_df = (
            spark.read.format("json")
            .option("multiline", "true")
            .option("encoding", encoding)
            .option("inferSchema", True)
            .option("forceLowercaseNames", True)
            .option("inferLong", True)
            .option("inferDecimal", True)
            .option("inferInteger", True)
            .load(
                ingress_file_path,
                forceLowercaseNames=True,
                inferLong=True,
                inferDecimal=True,
                inferInteger=True,
            )
        )

        # get column with the list of json for that type
        if len(source_json_path.strip()) > 0:
            # nested json
            list_columns = json_df.columns
            json_df = json_df.withColumn("json_column", to_json(list_columns[0]))
            json_df_to_pandas = json_df.toPandas()
            first_row = json_df_to_pandas["json_column"].iloc[0]
            json_data = json.loads(first_row)
            dict_json = cls.get_json_by_key(json_data, source_dataset_name)

            dict_list = dict_json.items()
            dict_dataframe = defaultdict(list)
            num_cols = 0
            for i, col_list in enumerate(dict_list):
                num_cols = len(col_list) + 1
                values = list(col_list)
                values.append(source_dataset_name)
                tuble_values = tuple(values)
                dict_dataframe[tuble_values].append(i)

            list_dataframe = list(dict_dataframe)

            new_cols = []
            for c_orig in range(1, num_cols):
                c_renamed = "column" + str(c_orig + 1)
                new_cols.append(c_renamed)
            new_cols.append("source_dataset_name")

            print(f"num_cols:{num_cols} new_cols:{new_cols}")
            print(f" dict_list:{str(dict_list)}")

            df_json = pd_legacy.DataFrame(list_dataframe, columns=new_cols)
            df_json.columns = new_cols
            print(df_json)
            unsorted_df = spark.createDataFrame(df_json)
            unsorted_df.na.fill(0)
            unsorted_df.na.fill("")
            unsorted_df.show()

        else:
            # flat json
            print(f"ingress_file_path:{ingress_file_path}")
            json_df_to_pandas = json_df.toPandas()
            json_string = json_df_to_pandas.to_json(orient="index")

            json_data = json.loads(json_string)
            dict_json = cls.get_json_by_key(json_data, "0")
            source_dataset_name = dataset_name
            dict_list = dict_json.items()
            dict_dataframe = defaultdict(list)
            num_cols = 0
            for i, x in enumerate(dict_list):
                num_cols = len(x) + 1
                values = list(x)
                values.append(source_dataset_name)
                tuble_values = tuple(values)
                dict_dataframe[tuble_values].append(i)

            list_dataframe = list(dict_dataframe)

            new_cols = []

            for c_orig in range(1, num_cols):
                c_renamed = "column" + str(c_orig + 1)
                new_cols.append(c_renamed)

            new_cols.append("source_dataset_name")

            print(f"num_cols:{num_cols} new_cols:{new_cols}")
            print(f" dict_list:{str(dict_list)}")

            df_json = pd_legacy.DataFrame(list_dataframe, columns=new_cols)
            df_json.columns = new_cols
            print(df_json)
            unsorted_df = spark.createDataFrame(df_json)
            unsorted_df.na.fill(0)
            unsorted_df.na.fill("")

        return unsorted_df

    # ingress_mount = f"/mnt/ddnid-nccdphp/OD/NHanes/dev/ingress"
    @classmethod
    def convert_sas_to_dataframe_with_schema(cls, spark: SparkSession, ingress_file_path: str, tenant_id: str,
                                             client_id: str, client_secret: str):
        """ Loads xpt or bdat to dataframe and extracts schema from xpt file
        - Reference article: https://stackoverflow.com/questions/58612304/reading-huge-sas-dataset-in-python
        - PS: Please be noted that the resulting dataframe allChunk is going to have all column as Categorical data

        Args:
            spark (SparkSession): SparkSession
            ingress_file_path (str): ADLS2 path to csv or usv file
            tenant_id (str): azure tenant id used to download file
            client_id (str): azure client id used to download file
            client_secret (str): azure secret used to download file

        Returns:
            DataFrame: Converted Spark dataframe, schema is extracted from xpt file
        """

        obj_env_file = pade_env_file.EnvironmentFile()
        https_path = obj_env_file.convert_abfss_to_https_path(ingress_file_path)
        credential = ClientSecretCredential(tenant_id, client_id, client_secret)

        # download file
        storage_account_loc = urlparse(https_path).netloc
        account_url = f"https://{storage_account_loc}"
        storage_path = urlparse(https_path).path
        storage_container = storage_path.split('/')[1]
        file_path = storage_path.replace(f"{storage_container}" + "/", "")
        print(f"storage_path:{storage_path}")
        print(f"https_path:{https_path}")
        # print(f"ingress_mount:{ingress_mount}")
        print(f"ingress_file_path:{ingress_file_path}")
        service_client = DataLakeServiceClient(account_url=account_url, credential=credential)
        file_system_client = service_client.get_file_system_client(storage_container)
        file_client = file_system_client.get_file_client(file_path)
        download = file_client.download_file(0)
        download_bytes = download.readall()
        download_file = BytesIO(download_bytes)

        # step 1: get pandas data frame

        chunk_size = 50000
        offset = 0
        # Get the function object in a variable getChunk
        if file_path.lower().endswith('sas7bdat'):
            get_chunk = prs.read_sas7bdat
        else:
            get_chunk = prs.read_xport

        df_all_chunks, _ = get_chunk(ingress_file_path, row_limit=chunk_size, row_offset=offset)
        df_all_chunks = df_all_chunks.astype('category')

        while True:
            offset += chunk_size
            # for xpt data, use pyreadstat.read_xpt()
            chunk, _ = prs.read_xport(ingress_file_path, row_limit=chunk_size, row_offset=offset)
            # if chunk is empty, it means the entire data has been read, so break
            if chunk.empty:
                break

            # converting each column to categorical
            for each_col in chunk:
                col_union = pd_legacy.api.types.union_categoricals([df_all_chunks[each_col], chunk[each_col]])
                df_all_chunks[each_col] = pd_legacy.Categorical(df_all_chunks[each_col],
                                                                categories=col_union.categories)
                chunk[each_col] = pd_legacy.Categorical(chunk[each_col], categories=col_union.categories)

            # Append each chunk to the resulting dataframe
            df_all_chunks = pd_legacy.concat([df_all_chunks, chunk])

        df_xpt, xpt_meta = prs.read_xport(download_file)
        df_data = spark.createDataFrame(df_xpt)

        # step 2: initalize empty pandas datafram
        df_metadata = pd.DataFrame()

        # read column name, labels into the new pandas dataframe
        df_metadata["name"] = xpt_meta.column_names = xpt_meta.column_names.str.strip()
        df_metadata["label"] = xpt_meta.column_labels = [x.decode('utf-8') for x in xpt_meta.column_labels]
        df_metadata["format"] = xpt_meta.column_formats = [x.decode('utf-8') for x in xpt_meta.column_formats]
        df_metadata["type"] = xpt_meta.column_types = [x.decode('utf-8') for x in xpt_meta.variable_types]
        df_metadata["type_in_source"] = xpt_meta.column_types = [x.decode('utf-8') for x in xpt_meta.original_variable_types]
        df_metadata["length"] = xpt_meta.column_lengths = xpt_meta.column_lengths.astype(int)
        df_metadata["note"] = xpt_meta.column_notes = [x.decode('utf-8') for x in xpt_meta.column_notes]
        df_metadata["table_name"] = xpt_meta.table_name = xpt_meta.table_name.decode('utf-8')

        df_metadata = spark.createDataFrame(df_metadata)
        df_metadata.show()
        # extracting number of rows from the Dataframe
        row_count = df_metadata.count()
        # extracting number of columns from the Dataframe
        col_count = len(df_metadata.columns)
        # printing
        print(f"Dimension of the Dataframe is: {(row_count, col_count)}")
        print(f"Number of Rows are: {row_count}")
        print(f"Number of Columns are: {col_count}")
        return df_data, df_metadata

    @classmethod
    def pandas_to_spark(cls, df_pandas, spark):
        """_summary_: Converts pandas dataframe to spark dataframe

        Args:
            df_pandas (_type_): _description_
            spark (_type_): _description_

        Returns:
            _type_: _description_
        """

        list_pandas=[]
        df_pandas = df_pandas.reset_index()  # make sure indexes pair with number of rows
        for index, row in df_pandas.iterrows():
            print(row['column_name'], row['comment'])
            row_to_append = (row['column_name'], row['comment'])
            list_pandas.append(row_to_append)
        rdd = spark.sparkContext.parallelize(list_pandas)
        rdd_columns = ["column_name","comment"]
        df_spark= rdd.toDF(rdd_columns)
        return df_spark

    @classmethod
    def convert_sas_mount_to_dataframe_with_schema(cls, spark: SparkSession, source_abbreviation: str, file_name: str, tenant_id: str,
                                             client_id: str, client_secret: str):
        """ Loads xpt or bdat to dataframe and extracts schema from xpt file
        - Reference article: https://stackoverflow.com/questions/58612304/reading-huge-sas-dataset-in-python
        - PS: Please be noted that the resulting dataframe allChunk is going to have all column as Categorical data

        Args:
            spark (SparkSession): SparkSession
            ingress_file_path (str): ADLS2 path to csv or usv file
            tenant_id (str): azure tenant id used to download file
            client_id (str): azure client id used to download file
            client_secret (str): azure secret used to download file

        Returns:
            DataFrame: Converted Spark dataframe, schema is extracted from xpt file
        """

        # obj_ds_core = pade_ds_core.DataSetCore()
        obj_env_file = pade_env_file.EnvironmentFile()
        mount_file_path = "/Workspace/Repos/tbi8@cdc.gov/pade/od/od_nhanes/temp/"
        ingress_path = f"{mount_file_path}{source_abbreviation}/{file_name}"
        print(f"mount_file_path:{mount_file_path}")
        print(f"ingress_path:{ingress_path}")

        # df_xpt, xpt_meta = prs.read_xport(ingress_path,metadataonly=True)
        df_xpt, xpt_meta = prs.read_xport(ingress_path)

        # column_name                    xpt_meta.column_names
        # data_type_name                 xpt_meta.variable_types
        # comment                        xpt_meta.column_notes.decode('utf-8')
        # dataset_name
        # full_dataset_name
        # dataset_file_path
        # row_id
        # unique_count
        # max_length
        # min_length
        # ingress_column_name            xpt_meta.column_names
        # ingress_column_format          xpt_meta.column_formats
        # ingress_column_label           xpt_meta.column_labels
        # unique_count_scrubbed
        # scope
        # row_id_column
        # row_count                      xpt_meta.number_rows.astype(int)
        # ingress_row_count              xpt_meta.number_rows.astype(int)
        # ingress_ordinal_position
        # ingress_column_length           xpt_meta.column_lengths.astype(int)
        # ingress_table_name              xpt_meta.table_name.decode('utf-8')


        list_pandas = list(zip(xpt_meta.column_names, xpt_meta.column_labels))
        print(f"list_pandas: {list_pandas}")
        rdd = spark.sparkContext.parallelize(list_pandas)

        schema = StructType(
            [
                StructField("column_name", StringType(), False),
                StructField("data_type_name", StringType(), False),
                StructField("comment", StringType(), True),
                StructField("dataset_name", StringType(), False),
                StructField("full_dataset_name", StringType(), False),
                StructField("dataset_file_path", StringType(), False),
                StructField("row_id", StringType(), False),
                StructField("unique_count", LongType(), True),
                StructField("null_count", LongType(), True),
                StructField("max_length", LongType(), True),
                StructField("min_length", LongType(), True),
                StructField("ingress_column_name", StringType(), True),
                StructField("ingress_column_format", StringType(), True),
                StructField("ingress_column_label", StringType(), True),
                StructField("unique_count_scrubbed", LongType(), True),
                StructField("scope", StringType(), True),
                StructField("row_id_column", StringType(), True),
            ]
        )

        schema_rdd = StructType([StructField("column_name", StringType(), False),
                                 StructField("column_label", StringType(), True)])
        df_metadata_spark = spark.createDataFrame(rdd, schema=schema_rdd)
        df_metadata_spark.printSchema()
        df_metadata_spark.show(truncate=False)

        # extracting number of rows from the Dataframe
        row = df_metadata_spark.count()
        # extracting number of columns from the Dataframe
        col = len(df_metadata_spark.columns)
        # printing
        print(f"Dimension of the Dataframe is: {(row,col)}")
        print(f"Number of Rows are: {row}")
        print(f"Number of Columns are: {col}")

        return df_metadata_spark, df_xpt


    @classmethod
    def convert_xpt_to_dataframe(cls, spark: SparkSession, ingress_file_path: str, tenant_id: str, client_id: str,
                                 client_secret: str) -> DataFrame:
        """Loads xpt to dataframe
        - XPT supports types - XPT types are converted to Spark types
        - Column names are not scrubbed or changed
        - No metadata is added to the dataframe
        - Pandas does not support accessing abfss directly so file is downloaded prior to loading

        Args:
            spark (SparkSession): SparkSession
            ingress_file_path (str): ADLS2 path to csv or usv file
            tenant_id (str): azure tenant id used to download file
            client_id (str): azure client id used to download file
            client_secret (str): azure secret used to download file

        Returns:
            DataFrame: Spark dataframe
        """

        obj_env_file = pade_env_file.EnvironmentFile()
        https_path = obj_env_file.convert_abfss_to_https_path(ingress_file_path)
        credential = ClientSecretCredential(tenant_id, client_id, client_secret)

        # download file
        storage_account_loc = urlparse(https_path).netloc
        account_url = f"https://{storage_account_loc}"
        storage_path = urlparse(https_path).path
        storage_container = storage_path.split('/')[1]
        file_path = storage_path.replace(f"{storage_container}" + "/", "")
        service_client = DataLakeServiceClient(account_url=account_url, credential=credential)
        file_system_client = service_client.get_file_system_client(storage_container)
        file_client = file_system_client.get_file_client(file_path)
        download = file_client.download_file(0)
        download_bytes = download.readall()
        download_file = BytesIO(download_bytes)

        # step 1: get pandas data frame
        df_xpt = pd_legacy.read_sas(download_file, format='xport')
        # step 2: get spark data frame
        unsorted_df = spark.createDataframe(df_xpt)

        return unsorted_df

    @classmethod
    def convert_sas_to_dataframe(cls, spark: SparkSession, ingress_file_path: str, tenant_id: str, client_id: str,
                                 client_secret: str) -> DataFrame:
        """Loads sas dta to dataframe
        - XPT supports types - XPT types are converted to Spark types
        - Column names are not scrubbed or changed
        - No metadata is added to the dataframe
        - Pandas does not support accessing abfss directly so file is downloaded prior to loading

        Args:
            spark (SparkSession): SparkSession
            ingress_file_path (str): ADLS2 path to csv or usv file
            tenant_id (str): azure tenant id used to download file
            client_id (str): azure client id used to download file
            client_secret (str): azure secret used to download file

        Returns:
            DataFrame: Spark dataframe
        """

        obj_env_file = pade_env_file.EnvironmentFile()
        https_path = obj_env_file.convert_abfss_to_https_path(ingress_file_path)
        credential = ClientSecretCredential(tenant_id, client_id, client_secret)

        # download file
        storage_account_loc = urlparse(https_path).netloc
        account_url = f"https://{storage_account_loc}"
        storage_path = urlparse(https_path).path
        storage_container = storage_path.split('/')[1]
        file_path = storage_path.replace(f"{storage_container}" + "/", "")
        service_client = DataLakeServiceClient(account_url=account_url, credential=credential)
        file_system_client = service_client.get_file_system_client(storage_container)
        file_client = file_system_client.get_file_client(file_path)
        download = file_client.download_file(0)
        download_bytes = download.readall()
        download_file = BytesIO(download_bytes)

        # step 1: get pandas data frame
        df_sas = pd_legacy.read_sas(download_file, format='sas7bdat')
        # step 2: get spark data frame
        unsorted_df = spark.createDataFrame(df_sas)
        return unsorted_df

    @classmethod
    def convert_xlsx_to_dataframe(cls, spark: SparkSession, ingress_file_path: str, sheet_name: str, tenant_id: str,
                                  client_id: str, client_secret: str,
                                  skip_rows_text: str) -> DataFrame:
        """Loads xlsx file to dataframe
        - Currently supports xlsx only not xls
        - Supports both formulas and values
        - Supports url extraction
        - All data is imported as strings
        - Original column names are stored in the metadata column
        - Column names are scrubbed for special characters and spaces

        Args:
            spark (SparkSession): SparkSession
            ingress_file_path (str): ADLS2 path to xlsx file
            sheet_name (str): name of the worksheet or ordinal position starting at 0
            tenant_id (str): azure tenant id used to download file
            client_id (str): azure client id used to download file
            client_secret (str): azure secret used to download file
            skip_rows (str): number of rows to skip at the top of the spreadsheet before headers

        Returns:
            DataFrame: Converted Spark dataframe
        """

        obj_ds_core = pade_ds_core.DataSetCore()
        print("attempting to load dataframe for xlsx")
        print(f"ingress_file_path:{ingress_file_path}")

        if skip_rows_text is None:
            skip_rows = 0
        else:
            if str(skip_rows_text) == "":
                skip_rows = 0
            elif len(str(skip_rows_text).strip()) == 0:
                skip_rows = 0
            else:
                skip_rows = int(skip_rows_text)

        if sheet_name is None:
            sheet_name = ""

        obj_env_file = pade_env_file.EnvironmentFile()
        https_path = obj_env_file.convert_abfss_to_https_path(ingress_file_path)
        credential = ClientSecretCredential(tenant_id, client_id, client_secret)

        # download file
        storage_account_loc = urlparse(https_path).netloc
        account_url = f"https://{storage_account_loc}"
        storage_path = urlparse(https_path).path
        storage_container = storage_path.split('/')[1]
        file_path = storage_path.replace(f"{storage_container}" + "/", "")
        service_client = DataLakeServiceClient(account_url=account_url, credential=credential)
        file_system_client = service_client.get_file_system_client(storage_container)
        file_client = file_system_client.get_file_client(file_path)
        download = file_client.download_file(0)
        download_bytes = download.readall()
        download_file = BytesIO(download_bytes)

        work_book_data = None
        print("try open file")
        print("try load workbook")
        print(f"file_path:{file_path}")

        if sheet_name is None or sheet_name == "":
            print("try read_excel")
            df_excel = pd_legacy.read_excel(
                download_file,
                skiprows=skip_rows,
                engine="openpyxl",
                dtype="string",
                na_filter=False,
                header=None,
            )
        else:
            sheet_name = str(sheet_name)
            print("try read_excel")
            if sheet_name.isdigit():
                df_excel = pd_legacy.read_excel(
                    download_file,
                    skiprows=skip_rows,
                    engine="openpyxl",
                    dtype="string",
                    sheet_name=int(sheet_name),
                    na_filter=False,
                    header=None,
                )
            else:
                df_excel = pd_legacy.read_excel(
                    download_file,
                    skiprows=skip_rows,
                    engine="openpyxl",
                    dtype="string",
                    sheet_name=sheet_name,
                    na_filter=False,
                    header=None,
                )

        data = df_excel.values.tolist()
        cols = np.asarray(data[0:1][0])
        data = data[skip_rows + 1:]
        print(f"original cols:{cols}")
        cols = cols.tolist()
        cols = obj_ds_core.rename_column_names_as_unique(cols)
        print(f"deduplicated cols:{cols}")
        print(f"len(cols):{len(cols)}")
        print("call loadworkbook data")
        print(f"2 open {https_path} work_book_data data_only False: read_only True")

        try:
            work_book_data = openpyxl.load_workbook(
                download_file, data_only=False, read_only=False
            )
            g_sheet = work_book_data.sheetnames
            if sheet_name is None or sheet_name == "":
                work_sheet_data = work_book_data[0]
            else:
                if sheet_name.isdigit():
                    work_sheet_data = work_book_data[g_sheet[int(sheet_name)]]
                else:
                    work_sheet_data = work_book_data[sheet_name]

            data_pass_1 = cls.convert_xlsx_to_list(data, cols, work_sheet_data, skip_rows)
            print("close work_book_data data_only True: read_only True")
        finally:
            if work_book_data is not None:
                work_book_data.close()

        # formula
        print(f"3 open {https_path} work_book_formula")
        print("data_only False: read_only False")
        work_book_formula = None
        try:
            work_book_formula = openpyxl.load_workbook(
                download_file, data_only=True, read_only=False
            )
            g_sheet = work_book_formula.sheetnames
            if sheet_name is None or sheet_name == "":
                work_sheet_formula = work_book_formula[0]
            else:
                if sheet_name.isdigit():
                    work_sheet_formula = work_book_formula[g_sheet[int(sheet_name)]]
                else:
                    work_sheet_formula = work_book_formula[sheet_name]

            print("call loadworkbook formula")
            data_pass_2 = cls.convert_xlsx_to_list(
                data_pass_1, cols, work_sheet_formula, skip_rows
            )

        finally:
            if work_book_formula is not None:
                work_book_formula.close()

        print("close work_book_formula data_only False: read_only False")
        # data = ( islice(r, i , None) for r in data)

        obj_ds_core = pade_ds_core.DataSetCore

        df_excel = pd_legacy.DataFrame(data_pass_2, columns=cols)
        df_excel = df_excel.astype(str)
        # df_excel.columns = df_excel.columns.str.replace(" ", "_")
        df_excel = df_excel.replace("nan", "")
        df_excel.replace(np.NaN, "", inplace=True)
        print(df_excel)
        unsorted_df = spark.createDataFrame(df_excel)
        for col_orig in unsorted_df.columns:
            # Data to replace format: Unnamed:_1
            c_renamed = obj_ds_core.scrub_object_name(col_orig)
            if c_renamed.startswith("unnamed:_"):
                c_renamed = c_renamed.replace("unnamed:_", "")
                # increments #by 1
                c_renamed = int(c_renamed) + 1
                c_renamed = "Column" + str(c_renamed)
            unsorted_df = unsorted_df.withColumnRenamed(col_orig, c_renamed)
            unsorted_df = unsorted_df.withMetadata(c_renamed, {"ingress_column_name": col_orig, 'comment': col_orig})

        # replace nulls
        unsorted_df.na.fill(0)
        unsorted_df.na.fill("")
        unsorted_df.show()
        unsorted_df = unsorted_df.withColumn("__meta_sheet_name", lit(sheet_name))

        return unsorted_df

    @staticmethod
    def convert_xlsx_to_list(data, cols, work_sheet, skip_rows):
        """Loads excel worksheet into a list of dictionaries

        Args:
            data (_type_): _description_
            cols (_type_): _description_
            work_sheet (_type_): _description_
            skip_rows (_type_): _description_

        Returns:
            list: Python list with converted Excel Data
        """

        # iterate through excel and display data
        for i_row, row_item in enumerate(data):
            row_item_updated = [""] * (len(cols))
            for j_col, col_item in enumerate(row_item):
                col_item_updated = col_item
                cell = work_sheet.cell(row=i_row + 2 + skip_rows, column=j_col + 1)
                if cell.hyperlink is not None:
                    url = str(cell.hyperlink.target)
                    col_item_updated = "[" + str(cell.value) + "](" + url + ")"
                else:
                    if cell.value is not None:
                        col_item_updated = cell.value
                    else:
                        col_item_updated = ""
                row_item_updated[j_col] = col_item_updated
                print(f"col_item_updated:{col_item_updated}")
            data[i_row] = row_item_updated
        return data



import requests
import time
import os
import sys
import json
from datetime import datetime
from dotenv import load_dotenv
from manifest import Manifest

def getAPIToken(alation_refresh_token, alation_user_id, alation_url):
    """Obtains an API token from Alation using a refresh token, user ID, and Alation URL.

    Args:
        alation_refresh_token (str): A refresh token obtained from Alation.
        alation_user_id (str): The user ID associated with the refresh token.
        alation_url (str): The URL of the Alation instance to connect to.

    Raises:
        Exception: If there is an error obtaining the API token.

    Returns:
        str: The API token.
    """

    print(f"Getting API token with {alation_refresh_token} refresh token for user {alation_user_id}")
    token_data = {
        "refresh_token": alation_refresh_token,
        "user_id": alation_user_id
    }
    alation_access_token = ""
    token_status = ""
    token_expires_at = None
    print('Token data is {token_data}')
    try:
        token_r = requests.post(
            '{base_url}/integration/v1/validateRefreshToken/'.format(base_url=alation_url
        ), data=token_data, verify = False, timeout = 30).json()
        token_status = token_r.get("token_status", "invalid").lower()
        token_expires_at = token_r.get("token_expires_at").split("T")[0]
    except Exception as e:
        print("Error in Alation refresh token validation request.")
        print("ERROR : "+str(e))
        raise e

    if token_status == "active":
        print("INFO: Alation Refresh token is valid")
        print("Token will expire on " + token_expires_at)
        # Regenerate token if expires within 7 days
        if token_expires_at:
            days_to_expiration = abs(datetime.strptime(token_expires_at, "%Y-%m-%d") - datetime.now()).days
            if days_to_expiration < 7:
                print('Alation Refresh Token will expire in ' + str(days_to_expiration) + ' days. Please create a new refresh token and replace the Pipeline API Token Variable.')
                sys.exit('Alation Refresh Token expiring in ' + str(days_to_expiration) + ' days.')
        
            try:
                access_token_r = requests.post(
                    '{base_url}/integration/v1/createAPIAccessToken/'.format(base_url=alation_url
                ), data=token_data, verify=False, timeout=30).json()
                alation_access_token = access_token_r.get("api_access_token")
                print('Alation API access token created is {alation_access_token}')
            except Exception as ex_access_token_request:
                print("Error in Alation access token request.")
                print(f"ERROR : {str(ex_access_token_request)}")
    elif token_status == "expired":
        print("ERROR: Alation Refresh Token has EXPIRED. Please create a new refresh token and replace the Pipeline API Token Variable.")
        sys.exit('Expired Alation Refresh Token.')
    else:
        print("ERROR: Alation Refresh Token is INVALID. Please create a new refresh token and replace the Pipeline API Token Variable.")
        sys.exit('Invalid Alation Refresh Token.')

    # 0.1 Create the Authorization headers with the API_TOKEN
    alation_headers = {
        "Accept": "application/json",
        "Content-Type": "application/json",
        "Token": alation_access_token,
        "token": alation_access_token
    }

    return alation_access_token, alation_headers


""" Module for spark and os environment for tech_environment_service with minimal dependencies. """

import os

# library management
from importlib import util  # library management

# error handling
from subprocess import check_output, Popen, PIPE, CalledProcessError
import subprocess

#  data
pyspark_pandas_loader = util.find_spec("pyspark.pandas")
pyspark_pandas_found = pyspark_pandas_loader is not None

class EnvironmentCore:
    """ EnvironmentCore class with minimal dependencies for the developer service.
    - This class is used to configure the python environment.
    - This class also provides a broad set of generic utility functions.
    """

    @classmethod
    def print_version(cls) -> str:
        """ Prints version of library

        Returns:
            str: version of library
        """

        print_version_command = ["poetry", "version"]
        print_version_command_string = ' '.join(print_version_command)
        print(print_version_command_string)
        current_working_dir = os. getcwd()
        print_version_result = f"current_working_dir:{current_working_dir}"
        try:
            print_version_result = check_output(print_version_command)
            # print_version_result = cls.execute(print_version_command)
            print_version_result = f"{str(print_version_result)}:{print_version_command_string} succeeded"
        except subprocess.CalledProcessError as ex_called_process:
            error_string = ex_called_process.output
            print_version_result = str(print_version_result)
            if error_string is None:
                new_error_string = f": {print_version_command_string} succeeded with Exception"
                print_version_result = print_version_result + new_error_string

            else:
                print_version_result = print_version_result + f"Error: {error_string}"

        print_version_result = str(print_version_result)
        print(print_version_result)
        return print_version_result



""" Module with a variety of metadata extraction routines for different source formats. """

# core
import sys  # don't remove required for error handling
import os

from urllib.parse import urlparse
from io import BytesIO

# libraries
from importlib import util

import pyreadstat as prs
import pandas as pd_legacy
pyspark_pandas_loader = util.find_spec("pyspark.pandas")
pyspark_pandas_found = pyspark_pandas_loader is not None
from pyspark.sql import SparkSession, DataFrame

if pyspark_pandas_found:
    import pyspark.pandas as pd
    # bug - pyspark version will not read local files
    # import pandas as pd
else:
    import pandas as pd

# adls and azure security
from azure.storage.filedatalake import DataLakeServiceClient
from azure.identity import ClientSecretCredential

# PADE
import pade_python.tech_environment_service.environment_file as pade_env_file

class DataSetExtract:
    """DataSetExtract class for Spark Datasets handle metadata extraction from different source formats
    """

    @classmethod
    def extract_xpt_dataframe_schema(cls, spark: SparkSession, ingress_file_path: str, tenant_id: str,
                                     client_id: str, client_secret: str) -> DataFrame:
        """ Extracts metadata from xpt file

        Args:
            spark (SparkSession): SparkSession
            ingress_file_path (str): ADLS2 path to csv or usv file
            tenant_id (str): azure tenant id used to download file
            client_id (str): azure client id used to download file
            client_secret (str): azure secret used to download file

        Returns:
            DataFrame: Spark dataframe
        """

        obj_env_file = pade_env_file.EnvironmentFile()
        https_path = obj_env_file.convert_abfss_to_https_path(ingress_file_path)
        credential = ClientSecretCredential(tenant_id, client_id, client_secret)

        # download file
        storage_account_loc = urlparse(https_path).netloc
        account_url = f"https://{storage_account_loc}"
        storage_path = urlparse(https_path).path
        storage_container = storage_path.split('/')[1]
        file_path = storage_path.replace(f"{storage_container}" + "/", "")
        service_client = DataLakeServiceClient(account_url=account_url, credential=credential)
        file_system_client = service_client.get_file_system_client(storage_container)
        file_client = file_system_client.get_file_client(file_path)
        download = file_client.download_file(0)
        download_bytes = download.readall()
        download_file = BytesIO(download_bytes)

        # step 1: get pandas data frame

        xpt_meta = prs.read_xport(download_file, metadataonly=True)

        # step 2: initalize empty pandas dataframe
        df_metadata = pd.DataFrame()

        # read column name, labels into the new pandas dataframe
        df_metadata["name"] = xpt_meta.column_names = xpt_meta.column_names.str.strip()
        df_metadata["label"] = xpt_meta.column_labels = [x.decode('utf-8') for x in xpt_meta.column_labels]
        df_metadata["format"] = xpt_meta.column_formats = [x.decode('utf-8') for x in xpt_meta.column_formats]
        df_metadata["type"] = xpt_meta.column_types = [x.decode('utf-8') for x in xpt_meta.variable_types]
        df_metadata["type_in_source"] = xpt_meta.column_types = [x.decode('utf-8') for x in xpt_meta.original_variable_types]
        df_metadata["length"] = xpt_meta.column_lengths = xpt_meta.column_lengths.astype(int)
        df_metadata["note"] = xpt_meta.column_notes = [x.decode('utf-8') for x in xpt_meta.column_notes]
        df_metadata["table_name"] = xpt_meta.table_name = xpt_meta.table_name.decode('utf-8')

        metadata_df = spark.createDataFrame(df_metadata)

        return metadata_df



""" Module with a variety of create, read, update and delete
functions for Spark data frames. """

import sys  # don't remove required for error handling
import os


from pathlib import Path
from urllib.parse import urlparse
from io import BytesIO

# types
import json
from collections import defaultdict

# libraries
from importlib import util

# util
import hashlib

# adls and azure security
from azure.identity import DefaultAzureCredential, ClientSecretCredential
from azure.storage.filedatalake import DataLakeServiceClient

# file excel
import openpyxl
import openpyxl.utils.cell

# PADE
import pade_python.tech_environment_service.dataset_core as pade_ds_core
import pade_python.tech_environment_service.environment_file as pade_env_file
import pade_python.tech_environment_service.security_core as pade_sec_core
import pade_python.tech_environment_service.environment_core as pade_env_core

# spark / data
import uuid
from pyspark.sql import SparkSession, DataFrame
import pyspark.sql.utils
import pyspark.sql.functions as f
from pyspark.sql.functions import (
    col,
    lit,
    concat_ws,
    udf
)

from pyspark.sql.types import StructType, StructField, LongType, StringType
uuid_udf = udf(lambda: str(uuid.uuid4()), StringType())

from delta.tables import DeltaTable
pyspark_pandas_loader = util.find_spec("pyspark.pandas")
pyspark_pandas_found = pyspark_pandas_loader is not None

if pyspark_pandas_found:
    import pyspark.pandas as pd
    # bug - pyspark version will not read local files in the repo
    # import pandas as pd
else:
    import pandas as pd


class DataSetCrud:
    """DataSetCrud class for Spark Datasets handling create, read, update and delete operations
    """

    @classmethod
    def upsert(
        cls,
        spark,
        config,
        dbutils,
        df_crud,
        full_dataset_name,
        dataset_file_path,
        is_using_dataset_folder_path_override,
        file_format,
        source_abbreviation,
        ingress_file_path,
        is_drop,
        partition_by,
        incremental
    ):
        """Inserts a record if it does not exist or updates the record if it exists. Stores the Dataframe as
        Delta dataset if the path is empty or tries to merge the data if found

        Args:
            spark (_type_): _description_
            config (_type_): _description_
            dbutils (_type_): _description_
            df_crud (_type_): _description_
            full_dataset_name (_type_): _description_
            dataset_file_path (_type_): _description_
            is_using_dataset_folder_path_override (bool): _description_
            file_format (_type_): _description_
            source_abbreviation (_type_): _description_
            is_drop (bool, optional): _description_. Defaults to False.
            partition_by (_type_, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_
        """

        obj_ds_core = pade_ds_core.DataSetCore()
        obj_env_file = pade_env_file.EnvironmentFile()

        yyyy_param = config["yyyy"]
        mm_param = config["mm"]
        dd_param = config["dd"]

        if yyyy_param is None:
            yyyy_param = ""

        if mm_param is None:
            mm_param = ""

        if dd_param is None:
            dd_param = ""

        dataset_name_list = full_dataset_name.split(".")
        if dataset_name_list[0] is not None:
            database_name = dataset_name_list[0]
        else:
            database_name = None

        if dataset_name_list[1] is not None:
            dataset_name = dataset_name_list[1]
        else:
            dataset_name = None

        if file_format == "delta":
            file_format = "parquet_delta"

        is_internal_dataset = True
        test_flag = ""
        location_flag = "managed internally"

        pade_database_folder = config["pade_database_folder"]
        not_found = -1

        if dataset_file_path.find(pade_database_folder) is not_found:
            is_using_dataset_folder_path_override = True
            is_internal_dataset = False
            test_flag = " not"
            location_flag = "managed externally"
        elif (is_using_dataset_folder_path_override is True and file_format == "parquet_delta"):
            is_internal_dataset = False
            test_flag = " not"
            location_flag = "managed externally"

        table_exists_flag = obj_ds_core.table_exists(spark, dataset_name, database_name)

        print(f"database_name:{database_name} dataset_name:{dataset_name}")
        print(f"location_flag: {location_flag}")
        print(f"pade_database_folder:{pade_database_folder}")
        print(f"test_flag:{test_flag} in {dataset_file_path}")
        print(f"test if dataset exists:{dataset_name}")
        print(f"for database: {database_name} returned: {table_exists_flag}")
        print(f"partition_by:{partition_by}")
        print(f"is_internal_dataset:{is_internal_dataset}")
        print(f"is_drop:{is_drop}")
        print(f"file_format:{file_format}")
        print(f"incremental:{incremental}")

        if is_internal_dataset is False and file_format == "parquet_delta":
            if obj_env_file.file_exists(config, dataset_file_path, dbutils):
                sql_command = "DROP TABLE IF EXISTS " + full_dataset_name
                spark.sql(sql_command)
                sql_command = f"CREATE TABLE  {full_dataset_name}  USING DELTA LOCATION '{dataset_file_path}'"
                print(f"attempting sql_command:{sql_command}")
                spark.sql(sql_command)
                print(f"created Delta dataset {full_dataset_name}:")
                print(f"at {dataset_file_path}")
            else:
                print("error attempting to load a dataset file that does not exist")
                print(f"or is internal: {dataset_file_path}")

        # dataset is parquet internal to database directory and full refresh
        if (is_internal_dataset is True and incremental == "replace"):
            print("upsert for dataset")
            print("upsert is_drop: " + str(is_drop))
            print("upsert file_format: " + file_format)
            print("upsert partition_by: " + str(partition_by))
            # Consider adding delete option later but will need to support schema
            # changes and will most likely have a performance hit
            # Delete will slow down code and increase maintenance but may provide better schema change logging
            # sql_command = f"DROP TABLE IF EXISTS {full_dataset_name}"
            # sql_command = f"DELETE FROM  {full_dataset_name}"
            # spark.sql(sql_command)
            # Consider adding delete option later but will need to support schema
            # changes and will most likely have a performance hit
            # Delete will slow down code and increase maintenance but may provide better schema change logging
            # sql_command = f"DELETE FROM  {full_dataset_name}"

            # if file_format != "parquet_delta":
            # print("not parquet")
            if partition_by is not None:
                partition_by_array = partition_by.split(",")
                print(f"writing {full_dataset_name} to")
                print(f"{dataset_file_path} with partition")
                print(f"by {partition_by}")
                one = 1
                # don't merge schema for replace
                # .option( "mergeSchema", "true")
                if len(partition_by_array) is one:
                    print("paritioned (1) parquet saveastable : managed folder")
                    p_by = partition_by_array[0].strip()
                    df_crud.write.mode("overwrite").format("delta").partitionBy(p_by).saveAsTable(full_dataset_name)
                else:
                    if len(partition_by_array) > one:
                        p_by_0 = partition_by_array[0].strip()
                        p_by_1 = partition_by_array[1].strip()
                        df_crud.write.mode("overwrite").format("delta").partitionBy(
                            p_by_0).partitionBy(
                            p_by_1).saveAsTable(full_dataset_name)
                    else:
                        # no partition
                        # don't merge schema for replace
                        # .option("mergeSchema", "true")
                        print("paritioned (0) parquet saveastable : non managed folder override")
                        df_crud.write.mode("overwrite").format("delta").saveAsTable(full_dataset_name)
            else:
                print("unparitioned parquet saveastable : managed folder")
                # .option("mergeSchema", "true")
                try:
                    df_crud.write.mode("overwrite").format("delta").saveAsTable(full_dataset_name)
                except pyspark.sql.utils.AnalysisException as ex_analysis:
                    print(f"Error saving dataframe: {full_dataset_name}")
                    print("An exception occurred: " + str(ex_analysis))
                    # rename and try again
                    # .option("mergeSchema", "true")
                    sql_command = f"DROP TABLE IF EXISTS {full_dataset_name}"
                    spark.sql(sql_command)
                    try:
                        df_crud.write.mode("overwrite").format("delta").saveAsTable(full_dataset_name)
                    except pyspark.sql.utils.AnalysisException as ex_analysis_1:
                        print("Error saving dataframe: full_dataset_name")
                        print("An exception occurred: " + str(ex_analysis_1))
                        for c_original in df_crud.columns:
                            c_renamed = obj_ds_core.scrub_object_name(c_original)
                            df_crud = df_crud.withColumnRenamed(c_original, c_renamed)

        # dataset is parquet internal to database directory and incremental refresh
        print(f"is_internal_dataset:{is_internal_dataset}")
        print(f"incremental:{incremental}")
        if (is_internal_dataset is True) and (incremental == "incremental" or incremental == "incremental_with_purge"):
            table_exists = obj_ds_core.table_exists(spark, dataset_name, database_name)
            is_external_table = False
            if (file_format != "parquet_delta" or dataset_name == "bronze_clc_schema"):
                is_external_table = True
            if table_exists and is_external_table:
                if (incremental == "incremental_with_purge" and dataset_name != "bronze_clc_schema"):
                    sql_command = f"DELETE FROM {full_dataset_name} where "
                    sql_command = sql_command + " __meta_ingress_file_path = '{ingress_file_path}'"
                    print(f"attempting sql_command:{sql_command}")
                    spark.sql(sql_command)

                print(f"attempting to modifying existing dataset:{full_dataset_name}")
                print(f"with data from {ingress_file_path} in format {file_format}")
                print(f"with merge at path {dataset_file_path}")

                match_expr = "delta.row_id = updates.row_id and updates.row_id = delta.row_id"
                delta_dataset = None

                try:
                    delta_dataset = DeltaTable.forPath(spark, dataset_file_path)

                    delta_dataset.alias("delta").merge(
                        df_crud.alias("updates"), match_expr
                    ).whenMatchedUpdateAll().whenNotMatchedInsertAll().execute()
                    return True
                except Exception as exception_check:
                    print("merge request input of the following dataframe:" + str(exception_check))
                    # display(df_crud)
                    print("merge request output of the following dataframe" + str(exception_check))
                    if delta_dataset is not None:
                        df_crud_history = delta_dataset.history()
                        # display(df_crud_history)
                    raise
            else:
                print(f"table_exists:{obj_ds_core.table_exists(spark, dataset_name, database_name)}")
                print(f"is_internal_dataset:{is_internal_dataset} dataset_name:{dataset_name}")
                print(f"database_name:{database_name} file_format:{file_format}")
                print("attempting to create Delta dataset")
                print(f"{full_dataset_name}: at")
                print(f"{dataset_file_path}")
                # Dataset hot references external file
                if is_internal_dataset is False:
                    if file_format == "parquet":
                        sql_command = f"CREATE TABLE IF NOT EXISTS {full_dataset_name} "
                        sql_command = sql_command + f" USING PARQUET LOCATION '{dataset_file_path}'"
                    else:
                        sql_command = f"CREATE TABLE IF NOT EXISTS {full_dataset_name} "
                        sql_command = sql_command + f" USING DELTA LOCATION '{dataset_file_path}'"
                    spark.sql(sql_command)
                    print(f"created Delta dataset {full_dataset_name}:")
                    print(f"at {dataset_file_path}")
                else:
                    if file_format == "parquet":
                        print("error attempting to load a parquet directory")
                        print(f"that does not exist: {dataset_file_path}")
                    else:
                        if partition_by is not None:
                            partition_by_array = partition_by.split(",")
                            print(f"writing {full_dataset_name} to {dataset_file_path}")
                            print(f"with partition by {partition_by}")

                            if len(partition_by_array) == 1:
                                print(f"df_crud.write.format('delta').partitionBy({partition_by_array[0].strip()})")
                                print(f".saveAsTable({full_dataset_name})")
                                df_crud.write.format("delta").option(
                                    "mergeSchema", "true"
                                ).partitionBy(
                                    partition_by_array[0].strip()
                                ).saveAsTable(
                                    full_dataset_name
                                )
                            else:
                                if len(partition_by_array) > 1:
                                    print(f"df_crud.write.format('delta').partitionBy({partition_by_array[0].strip()})")
                                    print(f".partitionBy({partition_by_array[1].strip()}).")
                                    print(f"saveAsTable({full_dataset_name})")
                                    df_crud.write.format("delta").option(
                                        "mergeSchema", "true"
                                    ).partitionBy(
                                        partition_by_array[0].strip()
                                    ).partitionBy(
                                        partition_by_array[1].strip()
                                    ).saveAsTable(
                                        full_dataset_name
                                    )
                                else:
                                    message_text = f"df_crud.write.format('delta').saveAsTable({full_dataset_name})"
                                    print(message_text)
                                    df_crud.write.format("delta").option(
                                        "mergeSchema", "true"
                                    ).saveAsTable(full_dataset_name)
                        else:
                            print(f"writing {full_dataset_name} to ")
                            print(f"{dataset_file_path} without partition")
                            df_crud.write.format("delta").option(
                                "mergeSchema", "true"
                            ).saveAsTable(full_dataset_name)

        return "success"

    @classmethod
    def get_export_dataset_or_view_schema_config(
        cls,
        config,
        config_dataset_or_notebook,
        spark,
        sorted_df,
        view_or_schema,
        schema_df
    ):
        """Save metata data of dataset of view to delta lake
        Creates metadata for every column on the dataframe
        Creates a summary row with the column name "all" to summary metrics for the entire dataframe

        Args:
            config (_type_): _description_
            config_dataset_or_notebook (_type_): _description_
            spark (_type_): _description_
            dbutils (_type_): _description_
            sorted_df (_type_): _description_
            view_or_schema (_type_): _description_

        Returns:
            _type_: _description_
        """

        obj_ds_core = pade_ds_core.DataSetCore()

        schema_dataset_file_path = config["schema_dataset_file_path"]
        row_id_column_names = ""
        pade_database_folder = config["pade_database_folder"]
        environment = config["environment"]
        project_id = config["project_id"]

        is_using_dataset_folder_path_override = False

        if view_or_schema == "view":
            dataset_name = config_dataset_or_notebook["view_name"]
            full_dataset_name = config_dataset_or_notebook["full_view_name"]
            dataset_file_path = "n_a"
        else:
            dataset_name = config_dataset_or_notebook["dataset_name"]
            full_dataset_name = config_dataset_or_notebook["full_dataset_name"]
            dataset_file_path = config_dataset_or_notebook["dataset_file_path"]

        row_id_keys = config_dataset_or_notebook["row_id_keys"]
        sorted_df.createOrReplaceTempView("dataset_sorted_df")

        override = config["is_export_schema_required_override"]
        pade_database_name = config['pade_database_name']
        is_export_schema_required_override = override
        schema_full_dataset_name = f"pade_{project_id}_{environment}.bronze_clc_schema"

        if is_export_schema_required_override != "force_off":
            clc_dataset_name = "bronze_clc_schema"
            if obj_ds_core.table_exists(spark, clc_dataset_name, pade_database_name) is True:
                print(f"delete {full_dataset_name} from: {clc_dataset_name}")
                delete_sql = f"Delete from {pade_database_name}.bronze_clc_schema"
                delete_sql = delete_sql + f" where full_dataset_name = '{full_dataset_name}'"
                df_deleted = spark.sql(delete_sql)
                print("deleted rows: " + str(df_deleted.count()))
            print(f"describe {full_dataset_name} for: bronze_clc_schema")
            df_schema = spark.sql("Describe Table dataset_sorted_df")
            df_schema = df_schema.distinct()
            df_schema = df_schema.withColumn("dataset_name", lit(dataset_name))
            df_schema = df_schema.withColumn("full_dataset_name", lit(full_dataset_name))
            df_schema = df_schema.withColumn("dataset_file_path", lit(dataset_file_path))
            df_schema = df_schema.withColumnRenamed("col_name", "column_name")
            df_schema = df_schema.withColumnRenamed("data_type", "data_type_name")
            row_id_keys_databricks = "col('dataset_name'),col('column_name')"
            arg_list = [eval(col_name.strip()) for col_name in row_id_keys_databricks.split(",")]
            df_schema = df_schema.withColumn("row_id_databricks", concat_ws("-", *arg_list))
            if row_id_keys is None:
                row_id_keys = ""
            col_list = [
                ((x.strip().replace("col('", "").replace("')", "")))
                for x in row_id_keys.split(",")
            ]
            row_id_column_names = str(",".join(col_list))
            print("row_id_column_names: " + row_id_column_names)
            print("updated databricks metadata for: schema_databricks_df")

            merged_df = df_schema
            merged_df = merged_df.withColumn("row_id", col("row_id_databricks"))
            merged_df = merged_df.drop("row_id_databricks")
            merged_df = merged_df.drop("row_id_koalas")
            merged_df = merged_df.withColumn("unique_count", lit(0))
            merged_df = merged_df.withColumn("null_count", lit(0))
            merged_df = merged_df.withColumn("max_length", lit(0))
            merged_df = merged_df.withColumn("min_length", lit(0))
            merged_df = merged_df.withColumn("ingress_column_name", lit(""))
            merged_df = merged_df.withColumn("ingress_column_format", lit(""))
            merged_df = merged_df.withColumn("ingress_column_label", lit(""))
            merged_df = merged_df.withColumn("unique_count_scrubbed", lit(0))
            merged_df = merged_df.withColumn("scope", lit("column"))
            merged_df = merged_df.withColumn("row_id_column", lit(row_id_column_names))
            merged_df = merged_df.withColumn("row_count",  lit(0))
            merged_df = merged_df.withColumn("ingress_row_count",  lit(0))
            merged_df = merged_df.withColumn("ingress_ordinal_position",  lit(0))
            merged_df = merged_df.withColumn("ingress_column_length",  lit(0))
            merged_df = merged_df.withColumn("ingress_table_name",  lit(0))

            schema = StructType(
                [
                    StructField("column_name", StringType(), False),
                    StructField("data_type_name", StringType(), False),
                    StructField("comment", StringType(), True),
                    StructField("dataset_name", StringType(), False),
                    StructField("full_dataset_name", StringType(), False),
                    StructField("dataset_file_path", StringType(), False),
                    StructField("row_id", StringType(), False),
                    StructField("unique_count", LongType(), True),
                    StructField("null_count", LongType(), True),
                    StructField("max_length", LongType(), True),
                    StructField("min_length", LongType(), True),
                    StructField("ingress_column_name", StringType(), True),
                    StructField("ingress_column_format", StringType(), True),
                    StructField("ingress_column_label", StringType(), True),
                    StructField("unique_count_scrubbed", LongType(), True),
                    StructField("scope", StringType(), True),
                    StructField("row_id_column", StringType(), True),
                    StructField("row_count", LongType(), True),
                    StructField("ingress_row_count", LongType(), True),
                    StructField("ingress_ordinal_position", LongType(), True),
                    StructField("ingress_column_length", LongType(), True),
                    StructField("ingress_table_name", StringType(), True),
                ]
            )

            # display(merged_df)

            schema_column_df = spark.createDataFrame(merged_df.rdd, schema=schema)
            key = "is_using_dataset_folder_path_override"
            is_using_dataset_folder_path_override = config[key]
            schema_dataset_file_path = pade_database_folder.rstrip("/") + "/bronze_clc_schema"
            schema_dataset_file_path = schema_dataset_file_path.replace("dbfs:", "")
            partitioned_by = "dataset_name"

            empty = ""
            row_id = full_dataset_name
            scope = "dataset"
            row_data = [
                (
                    ("all_columns"),
                    "n_a",
                    "",
                    dataset_name,
                    full_dataset_name,
                    dataset_file_path,
                    row_id,
                    (0),
                    (0),
                    (0),
                    (0),
                    (empty),
                    (empty),
                    (empty),
                    (0),
                    (scope),
                    "full_dataset_name",
                     (0),
                     (0),
                     (0),
                     (0),
                     (0),
                )
            ]

            print("row_data: " + str(row_data))
            schema_dataset_df = spark.createDataFrame(row_data, schema)
        else:
            partitioned_by = ""
            schema_column_df = None
            schema_dataset_df = None

        config_schema = {
            "schema_column_df": schema_column_df,
            "schema_dataset_df": schema_dataset_df,
            "schema_full_dataset_name": schema_full_dataset_name,
            "schema_dataset_file_path": schema_dataset_file_path,
            "partitioned_by": partitioned_by,
            "is_using_dataset_folder_path_override": is_using_dataset_folder_path_override,
        }

        return config_schema



""" Module for python logging for tech_environment_service with minimal dependencies. """

from opencensus.ext.azure.log_exporter import AzureLogHandler
import sys  # don't remove required for error handling
import os
import logging
import logging.config

# Import from sibling directory ..\tech_environment_service
OS_NAME = os.name

sys.path.append("..")
if OS_NAME.lower() == "nt":
    print("windows")
    sys.path.append(os.path.dirname(os.path.abspath(__file__ + "\\..")))
    sys.path.append(os.path.dirname(os.path.abspath(__file__ + "\\..\\..")))
    sys.path.append(os.path.dirname(os.path.abspath(__file__ + "\\..\\..\\..")))
    env_path = os.path.dirname(os.path.abspath(sys.executable + "\\.."))
    env_share_path = env_path + "\\share"
    sys.path.append(os.path.dirname(os.path.abspath(sys.executable + "\\..\\share")))
    ERROR_LOG_FILENAME = env_share_path + "\\.pade_python_services_errors.log"
else:
    print("non windows")
    sys.path.append(os.path.dirname(os.path.abspath(__file__ + "/..")))
    sys.path.append(os.path.dirname(os.path.abspath(__file__ + "/../..")))
    sys.path.append(os.path.dirname(os.path.abspath(__file__ + "/../../..")))
    env_path = os.path.dirname(os.path.abspath(sys.executable + "/.."))
    env_share_path = env_path + "/share"
    sys.path.append(os.path.dirname(os.path.abspath(sys.executable + "/../share")))
    ERROR_LOG_FILENAME = env_share_path + "/.pade_python_services_errors.log"

FOLDER_EXISTS = os.path.exists(env_share_path)
if not FOLDER_EXISTS:
    # Create a new directory because it does not exist
    os.makedirs(env_share_path)


print(f"Log files stored at ERROR_LOG_FILENAME:{ERROR_LOG_FILENAME}")

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "format": "%(asctime)s:%(name)s:%(process)d:%(lineno)d " "%(levelname)s %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
        },
        "simple": {
            "format": "%(message)s",
        },
    },
    "handlers": {
        "logfile": {
            "class": "logging.handlers.RotatingFileHandler",
            "level": "ERROR",
            "filename": ERROR_LOG_FILENAME,
            "formatter": "default",
            "backupCount": 2,
        },
        "verbose_output": {
            "class": "logging.StreamHandler",
            "level": "DEBUG",
            "formatter": "simple",
            "stream": "ext://sys.stdout",
        },
    },
    "loggers": {
        "tryceratops": {
            "level": "INFO",
            "handlers": [
                "verbose_output",
            ],
        },
    },
    "root": {"level": "DEBUG", "handlers": ["logfile"]},
    "ddt_ops_dev": {"level": "DEBUG", "handlers": ["logfile"]},
    "ndsp_pertussis_dev": {"level": "DEBUG", "handlers": ["logfile"]}
}

class EnvironmentLogging:
    """ EnvironmentLogging class with minimal dependencies for the developer service.
    - This class is used to perform file and directory operations.
    """

    @staticmethod
    def log_debug(project_id_env, message, instrumentation_key) -> str:
        """ Log debug message. """

        logging.config.dictConfig(LOGGING_CONFIG)
        logger = logging.getLogger(project_id_env)
        logger.addHandler(AzureLogHandler(
            connection_string=f'InstrumentationKey={instrumentation_key}')
        )
        logger.debug(message)

        return "Success"

    @staticmethod
    def log_info(project_id_env, message, instrumentation_key) -> str:
        """ Log info message. """

        logging.config.dictConfig(LOGGING_CONFIG)
        logger = logging.getLogger(project_id_env)
        logger.addHandler(AzureLogHandler(
            connection_string=f'InstrumentationKey={instrumentation_key}')
        )

        logger.info(message)

        return "Success"

    @staticmethod
    def log_warning(project_id_env, message, instrumentation_key) -> str:
        """ Log warning message. """

        logging.config.dictConfig(LOGGING_CONFIG)
        logger = logging.getLogger(project_id_env)
        logger.addHandler(AzureLogHandler(
            connection_string=f'InstrumentationKey={instrumentation_key}')
        )
        logger.warning(message)

        return "Success"

    @staticmethod
    def log_error(project_id_env, message, instrumentation_key) -> str:
        """ Log error message. """

        logging.config.dictConfig(LOGGING_CONFIG)
        logger = logging.getLogger(project_id_env)
        logger.addHandler(AzureLogHandler(
            connection_string=f'InstrumentationKey={instrumentation_key}')
        )
        logger.error(message)

        return "Success"



""" Module for security_core for security_service with minimal dependencies. """

# core
from distutils.log import error
import os
import sys  # don't remove required for error handling
sys.path.append(".")

# errors
import traceback  # don't remove required for error handling

# text
import json
from html.parser import HTMLParser  # web scraping html

# http
import requests

# security
from adal import AuthenticationContext

import pade_python.developer_service.environment_logging as pade_env_log

class SecurityCore:
    """ Security functions with minimal dependencies
    """

    @classmethod
    def acquire_access_token_with_client_credentials(cls, sp_client_id: str, sp_client_secret: str, sp_tenant_id: str,
                                                     sp_redirect_url: str, sp_authority_host_url: str,
                                                     sp_azure_databricks_resource_id: str, project_id: str) -> dict:
        """Takes in config dictionary, client_id and client secret and returns config_user with access_token
        - initial call

        Args:
            config (dict): global config dictionary
            sp_client_id (str): service principal client id
            sp_client_secret (str): service principal secret
            sp_tenant_id (str): service principal tenant id
            sp_redirect_url (str): service principal redirect url
            sp_azure_databricks_resource_id (str): service principal azure databricks resource id
            project_id(str): project id for logging

        Returns:
            dict: config_user dictionary with access_token populated
        """

        info_message = f"acquire_access_token_with_client_credentials for {project_id}"
        print(info_message)

        config_user = cls.setup_user_configuration(sp_client_id, sp_client_secret, sp_tenant_id, sp_redirect_url,
                                                   sp_authority_host_url, sp_azure_databricks_resource_id)

        authority_url = config_user["authority_url"]
        azure_databricks_resource_id = config_user["azure_databricks_resource_id"]
        print(f"authority_url:{authority_url}")
        print(f"sp_client_id:{sp_client_id}")
        print(f"azure_databricks_resource_id:{azure_databricks_resource_id}")
        context = AuthenticationContext(authority_url)

        token_response = {"accessToken": "not_set"}

        #resource='https://management.core.windows.net/'
        #resource = "https://database.windows.net/"

        # Get token using username password first
        try:
            token_response = context.acquire_token_with_client_credentials(
                azure_databricks_resource_id, sp_client_id, sp_client_secret)
        except Exception as ex_access_token_with_client_credentials:
            # Get current system exception
            ex_type, ex_value, ex_traceback = sys.exc_info()

            # Extract unformatter stack traces as tuples
            trace_back = traceback.extract_tb(ex_traceback)

            # Format stacktrace
            stack_trace = list()
            error_string = "Error: Unable to acquire_access_token_with_client_credentials with sp_client_id"
            error_string = error_string + f":{sp_client_id}: Details: {str(ex_access_token_with_client_credentials)}"
            error_string = error_string + ": Extended:"
            for trace in trace_back:
                error_string = f"File : {trace[0]} , Line : {trace[1]}, Func.Name : {trace[2]}"
                error_string = error_string + f", Message : {trace[3]}, Type : {ex_type}"
                error_string = error_string + f", Value : {ex_value}"
                stack_trace.append(error_string)
            print(error_string)
        # self.validate_token_response_username_password(token_response)

        print(f"token_response: [REDACTED]: length:{len(str(token_response))}")
        # Use returned refresh token to acquire a new token.
        access_token = token_response["accessToken"]
        config_user["access_token"] = access_token

        return config_user

    @classmethod
    def acquire_access_token_with_refresh_token(cls, sp_client_id: str, sp_client_secret: str, sp_tenant_id: str,
                                                sp_redirect_url: str, sp_authority_host_url: str,
                                                sp_azure_databricks_resource_id: str) -> dict:
        """Takes in config dictionary, client_id and client secret and returns config_user with access_token
        - refresh token call

        Args:
            config (dict): global config dictionary
            sp_client_id (str): service principal client id
            sp_client_secret (str): service principal secret

        Returns:
            dict: config_user with refresh_token populated
        """

        config_user = cls.setup_user_configuration(sp_client_id, sp_client_secret, sp_tenant_id, sp_redirect_url,
                                                   sp_authority_host_url, sp_azure_databricks_resource_id)

        if config_user is None:
            print("config_user is None")
        else:
            print(f"config_user exists: {str(config_user)}")

        authority_url = config_user["authority_url"]
        # client_id = config_user['client_id']
        azure_databricks_resource_id = config_user["azure_databricks_resource_id"]
        print(f"creating context for authority_url:{authority_url}")
        context = AuthenticationContext(authority_url)
        if context is None:
            print(f"AuthenticationContext: None : not acquired for authority_url:{authority_url}")
        elif len(str(context)) == 0:
            print(f"AuthenticationContext: empty string : not acquired for authority_url:{authority_url}")
        else:
            print(f"AuthenticationContext acquired for authority_url:{authority_url}")

        user_id = "zfi4@cdc.gov"  # TO DO MAKE Configurable
        print(f"attempting to acquire token for azure_databricks_resource_id:{azure_databricks_resource_id}")
        token_response = context.acquire_token(azure_databricks_resource_id, user_id, sp_client_id)

        if token_response is None:
            print(f"token_response not found:{str(token_response)}")
            config_user["refresh_token"] = "error"
        else:
            t_s = f"acquired token_response:{str(token_response)} for azure_databricks_resource_id"
            t_s = t_s + f":{str(azure_databricks_resource_id)}"
            refresh_token = token_response["refreshToken"]
            config_user["refresh_token"] = refresh_token

        config_user_result = cls.refresh_access_token(config_user)

        return config_user_result

    @staticmethod
    def setup_user_configuration(sp_client_id: str, sp_client_secret: str, sp_tenant_id: str, sp_redirect_url: str,
                                 sp_authority_host_url: str, sp_azure_databricks_resource_id: str) -> dict:
        """Takes in a config dictionary, client_id and client_secret and returns populated config_user dictionary

        Args:
            config (dict): global config dictionary
            sp_client_id (str): service principal client id
            sp_client_secret (str): service principal client secret
            sp_tenant_id (str): service principal tenant id
            sp_redirect_url (str): service principal redirect url
            sp_authority_host_url (str): service principal authority host url
            sp_azure_databricks_resource_id (str): service principal azure databricks resource id

        Returns:
            dict: populated config_user dictionary
        """

        sp_authority_url = "".join([sp_authority_host_url.rstrip("/"), "/", sp_tenant_id])

        # todo change from check if all exist rather than any
        # client_secret_exists = coalesce(sp_tenant_id, sp_redirect_url, sp_client_id, sp_client_secret)

        # if (client_secret_exists is None):
        #    client_secret_exists = False

        config_user = {
            "tenant": sp_tenant_id,
            "client_id": sp_client_id,
            "redirect_uri": sp_redirect_url,
            "client_secret": sp_client_secret,
            "authority_host_url": sp_authority_host_url,
            "azure_databricks_resource_id": sp_azure_databricks_resource_id,
            "authority_url": sp_authority_url,
        }

        # print(f"config_user:{str(config_user)}")
        return config_user

    @staticmethod
    def refresh_access_token(config_user: dict) -> dict:
        """Takes in config_user dictionary, returns config_user with access and refresh token

        Args:
            config_user (dict): config_user dictionary

        Returns:
            dict: config_user dictionary populated with with refresh token
        """

        authority_url = config_user["authority_url"]
        client_id = config_user["client_id"]
        client_secret = config_user["client_secret"]
        refresh_token = config_user["refresh_token"]
        azure_databricks_resource_id = config_user["azure_databricks_resource_id"]

        context = AuthenticationContext(authority_url)
        token_response = context.acquire_token_with_refresh_token(
            refresh_token, client_id, azure_databricks_resource_id, client_secret
        )

        refresh_token = token_response["refreshToken"]
        access_token = token_response["accessToken"]

        config_user["refresh_token"] = refresh_token
        config_user["access_token"] = access_token

        print(str("config_user:{config_user}"))

        return config_user

    @staticmethod
    def get_pat_tokens(config: dict, token: str):
        """Takes in a config dictionary, token and base_path, returns populated list of pat tokens

        Args:
            config (dict): global config dictionary
            token (str): token

        Returns:
            list: list of pat tokens
        """

        databricks_instance_id = config["databricks_instance_id"]
        headers = {"Authentication": f"Bearer {token}"}
        url = f"https://{databricks_instance_id}/api/2.0/preview/permissions/authorization/tokens"

        print(f"url:{str(url)}")
        headers_redacted = str(headers).replace(token, "[bearer REDACTED]")
        print(f"headers:{headers_redacted}")

        response = requests.get(url=url, headers=headers)
        data = None

        try:
            response_text = str(response.text)
            data = json.loads(response_text)
            msg = f"Received credentials with length : {len(str(response_text))} when posting to : "
            msg = msg + "{url}"
            response_text_fetch = msg
            print("- response : success  -")
            print(f"{response_text_fetch}")
            results = data["access_control_list"]

        except Exception as exception_object:
            f_filter = HTMLFilter()
            f_filter.feed(response.text)
            response_text = f_filter.text
            print(f"- response : error - {exception_object}")
            print(f"Error converting response text:{response_text} to json")
            results = []

        return results

    @staticmethod
    def get_credentials_git(config: dict, token: str):
        """Takes in a config dictionary, token and base_path, returns populated list of files

        Args:
            config (dict): global config dictionary
            token (str): token
            base_path (str): path to list files

        Returns:
            list: list of files at the path location
        """

        databricks_instance_id = config["databricks_instance_id"]
        headers = {"Authentication": f"Bearer {token}"}
        url = f"https://{databricks_instance_id}/api/2.0/git-credentials"

        print(f"url:{str(url)}")
        headers_redacted = str(headers).replace(token, "[bearer REDACTED]")
        print(f"headers:{headers_redacted}")

        response = requests.get(url=url, headers=headers)
        data = None

        try:
            response_text = str(response.text)
            data = json.loads(response_text)
            msg = f"Received credentials with length : {len(str(response_text))} when posting to : "
            msg = msg + "{url}"
            response_text_fetch = msg
            print("- response : success  -")
            print(f"{response_text_fetch}")
            results = data["credentials"]

        except Exception as exception_object:
            f_filter = HTMLFilter()
            f_filter.feed(response.text)
            response_text = f_filter.text
            print(f"- response : error - {exception_object}")
            print(f"Error converting response text:{response_text} to json")
            results = []

        return results

    @staticmethod
    def set_credentials_git(config: dict, token: str):
        """Takes in a config dictionary, token and base_path, returns populated list of files

        Args:
            config (dict): global config dictionary
            token (str): token
            base_path (str): path to list files

        Returns:
            list: list of files at the path location
        """

        databricks_instance_id = config["databricks_instance_id"]
        headers = {"Authentication": f"Bearer {token}"}
        url = f"https://{databricks_instance_id}/api/2.0/git-credentials"

        print(f"url:{str(url)}")
        headers_redacted = str(headers).replace(token, "[bearer REDACTED]")
        print(f"headers:{headers_redacted}")

        response = requests.get(url=url, headers=headers)
        data = None

        try:
            response_text = str(response.text)
            data = json.loads(response_text)
            msg = f"Received credentials with length : {len(str(response_text))} when posting to : "
            msg = msg + "{url}"
            response_text_fetch = msg
            print("- response : success  -")
            print(f"{response_text_fetch}")
            results = data["credentials"]

        except Exception as exception_object:
            f_filter = HTMLFilter()
            f_filter.feed(response.text)
            response_text = f_filter.text
            print(f"- response : error - {exception_object}")
            print(f"Error converting response text:{response_text} to json")
            results = []

        return results

class HTMLFilter(HTMLParser):
    """Parses HTMLData

    Args:
        HTMLParser (_type_): _description_
    """

    text = ""

    def handle_data(self, data):
        """Parses HTMLData

        Args:
            data (_type_): _description_
        """
        self.text += data



""" Module for job funciontality with minimal dependencies. """

import os
import sys  # don't remove required for error handling
from pathlib import Path
from datetime import date

import traceback  # don't remove required for error handling

class JobCore:
    """ JobCore class with minimal dependencies for the developer service.
    - This class is used to execute jobs.
    """

    @staticmethod
    def get_standard_parameters(environment: str, dbutils) -> dict:
        """Takes in dbutils and returns populated parameters dictionary

        Tries to parse virtual environment name for project name
        If virtual environment name does not have _ then the system
        looks at the current folder path to project name
        If the system can't use the environment or folder path then
        the system default to ddt_ops_dev

        Args:
            environment: Environment - if None then is extracted from directory or virtual environment
            dbutils (_type_): Databricks dbutils object

        Returns:
            dict: populated parameters dictionary
        """


        path = Path(os.getcwd())
        virtual_env = os.path.basename(os.path.normpath(sys.prefix))

        if len(virtual_env.rsplit('_', 1)) == 1:
            if len(path.parts) > 2:
                project_id = os.path.basename(os.path.normpath(path))
                virtual_env = project_id + "_dev"
            else:
                virtual_env = 'ddt_ops_dev'
                project_id = "ddt_ops"

            if environment is None or environment == "":
                environment = "dev"
        else:
            virtual_env = virtual_env.lower()
            project_id = virtual_env.rsplit('_', 1)[0]
            if environment is None or environment == "":
                environment = virtual_env.rsplit('_', 1)[1]

        project_id_root = project_id.split('_', 1)[0]
        project_id_individual = project_id.split('_', 1)[1]

        print("virtual_env: " + virtual_env)

        if dbutils is None:
            running_local = True
            repository_path = str(path.parent.parent)
            yyyy_param = str(date.today().year)
            mm_param = f"{format(date.today().month,'02')}"
            dd_param = ""
        else:
            report_yyyy_values = ["2021", "2022", str(date.today().year)]
            report_yyyy = str(date.today().year)
            report_mm_values = ["01", "02", "03", "04", "05", "06", "07", "08",
                                "09", "10", "11", "12"]
            report_mm = f"{format(date.today().month,'02')}"
            report_dd_values = ["NA", "01", "02", "03", "04", "05", "06", "07",
                                "08", "09", "10", "11", "12", "13", "14",
                                "15", "16", "17", "18", "19", "20", "21", "22",
                                "23", "24", "25", "26", "27", "28", "29",
                                "30", "31"]
            report_dd = "NA"

            widgets = dbutils.widgets
            if widgets is not None:
                widgets.dropdown("report_yyyy", report_yyyy, report_yyyy_values)
                widgets.dropdown("report_mm", report_mm, report_mm_values)
                widgets.dropdown("report_dd", report_dd, report_dd_values)

            running_local = False
            repository_path = str(path.parent.parent)
            yyyy_param = dbutils.widgets.get("report_yyyy")
            if yyyy_param is None:
                yyyy_param = str(date.today().year)

            mm_param = dbutils.widgets.get("report_mm")
            if mm_param is None:
                mm_param = f"{format(date.today().month,'02')}"

            dd_param = dbutils.widgets.get("report_dd")
            if dd_param is None:
                dd_param = f"{format(date.today().day,'02')}"

        azure_client_secret_key = "PADE_" + virtual_env.upper() + "_AZURE_CLIENT_SECRET"
        project_id_root = project_id.rsplit('_', 1)[0]
        project_id_individual = project_id.rsplit('_', 1)[1]

        dataset_name = 'all'
        cicd_action = 'pull_request'

        parameters = {
            "environment": environment,
            "project_id_root": project_id_root,
            "project_id_individual": project_id_individual,
            "project_id": project_id,
            "yyyy": yyyy_param,
            "mm": mm_param,
            "dd": dd_param,
            "repository_path": repository_path,
            "dataset_name": dataset_name,
            "cicd_action": cicd_action,
            "running_local": running_local,
            "azure_client_secret_key": azure_client_secret_key
        }

        return parameters



""" Module for spark and os environment for tech_environment_service with
 minimal dependencies. """

# library management
from importlib import util  # library management

# error handling
from subprocess import check_output, Popen, PIPE, CalledProcessError

import sys  # don't remove required for error handling
import os

# files
import glob
import fileinput
import codecs
import chardet

# error handling
import subprocess

# http
from urllib.parse import urlparse
import requests

import pade_python.tech_environment_service.repo_core as pade_repo

# azcopy and adls
from azure.identity import DefaultAzureCredential
from azure.identity import ManagedIdentityCredential
from azure.identity import ClientSecretCredential
from azure.storage.filedatalake import DataLakeServiceClient
from azure.storage.filedatalake import DataLakeDirectoryClient
from azure.keyvault.secrets import SecretClient

# spark
from pyspark.sql import (SparkSession)
from pyspark.sql.types import (IntegerType, LongType, StringType, StructField,
                               StructType)

#  data
pyspark_pandas_loader = util.find_spec("pyspark.pandas")
pyspark_pandas_found = pyspark_pandas_loader is not None

if pyspark_pandas_found:
    # import pyspark.pandas  as pd
    # bug - pyspark version will not read local files in the repo
    import pyspark.pandas as pd
else:
    import pandas as pd


class EnvironmentFile:
    """ EnvironmentFile class with minimal dependencies for the developer
    service.
    - This class is used to perform file and directory operations.
    """

    @staticmethod
    def class_exists() -> bool:
        """Basic check to make sure object is instantiated

        Returns:
            bool: true/false indicating object exists
        """
        return True

    @staticmethod
    def convert_abfss_to_https_path(abfss_path: str) -> str:
        """Converts abfs path to https path

        Args:
            abfss_path (str): abfss path

        Returns:
            str: https path
        """
        hostname = abfss_path.split('/')[2]
        file_system = hostname.split('@')[0]
        print(f"hostname:{hostname}")
        print(f"file_system:{file_system}")
        storage_account = hostname.split('@')[1]
        print(f"storage_account:{storage_account}")
        https_path = abfss_path.replace(hostname, storage_account + '/' + file_system)
        https_path = https_path.replace('abfss', 'https')
        return https_path

    @staticmethod
    def convert_windows_dir(folder_path: str) -> str:
        """Converts a window folder path to bash format

        Args:
            folder_path (str): path to convert

        Returns:
            str: _converted path
        """

        window_dir = "\\"

        if window_dir in folder_path:
            folder_path = folder_path.replace("\\", "/")
            folder_path = folder_path.lstrip("/") + "/"
        else:
            folder_path = folder_path.rstrip("/") + "/"

        return folder_path

    @staticmethod
    def execute(cmd) -> str:
        """Takes a command and executes it

        Args:
            cmd (_type_): Command to execute

        Raises:
            CalledProcessError: Error executing command

        Returns:
            str: output of command
        """

        result_string = ""

        with Popen(cmd,
                   stdout=PIPE,
                   bufsize=1,
                   universal_newlines=True) as p_output:
            if p_output is not None:
                stdout = p_output.stdout
                if stdout is not None:
                    for line in stdout:
                        # process line here
                        result_string = result_string + line
            else:
                result_string = "p_output is None"

        if p_output.returncode != 0:
            raise CalledProcessError(p_output.returncode, p_output.args)

        return result_string

    @staticmethod
    def download_file(url: str) -> str:
        """Downloads a file from a url
        May require refactoring if memory becomes an issue

        Args:
            url (str): url of the file to download

        Returns:
            str: local file name
        """
        local_filename = url.split('/')[-1]
        # NOTE the stream=True parameter below
        with requests.get(url, stream=True) as request_result:
            request_result.raise_for_status()
            with open(local_filename, 'wb') as file_result:
                for chunk in request_result.iter_content(chunk_size=8192):
                    # If you have chunk encoded response uncomment if
                    # and set chunk_size parameter to None.
                    # if chunk:
                    file_result.write(chunk)
        return local_filename

    @classmethod
    def prepend_line_to_file(cls, source_path: str, destination_path: str,
                             line_to_prepend: str) -> str:
        """Add line to the beginning of a file

        Args:
            source_path (str): _description_
            destination_path (str): _description_
            line_to_prepend (str): _description_

        Returns:
            str: Status of operation
        """

        result = "running"
        print(f"source_path: {source_path}")
        print(f"destination_path: {destination_path}")
        with open(source_path,
                  'r',
                  encoding='utf-8') as original:
            data = original.read()
        with open(destination_path,
                  'w',
                  encoding='utf-8') as modified:
            modified.write(f"{line_to_prepend}\n" + data)
        result = "Success"
        return result

    @classmethod
    def combine_files(cls, source_path: str, file_mask: str,
                      destination_path: str) -> str:
        """Joins/combines multilple files

        Args:
            source_path (str): _description_
            file_mask (str): _description_
            destination_path (str): _description_

        Returns:
            str: Status of operation
        """
        result = "running"
        source_files = f"{source_path}{file_mask}"
        all_files = glob.glob(source_files)
        with open(destination_path,
                  'w+',
                  encoding='utf-8',
                  newline='\n') as f_output:
            for filename in all_files:
                print(f"filename:{filename}")
                with open(filename,
                          'r',
                          encoding="utf-8",
                          newline='\n') as f_input:
                    for line in f_input:
                        f_output.write(line)
        result = "Success"
        return result

    @classmethod
    def rename_directory(cls, config: dict, source_path, new_directory_name) -> str:
        """Rename ABFSS directory

        Args:
            config (dict): _description_
            source_path (_type_): _description_
            new_directory_name (_type_): _description_

        Returns:
            str: rename_status
        """

        try:
            client_id = config['client_id']
            client_secret = config['client_secret']

            result = "file_adls_copy failed"

            if client_secret is None:
                azure_client_secret_key = str(config["azure_client_secret_key"])
                key = azure_client_secret_key
                client_secret = f"Environment variable: {key} not found"

            os.environ['AZCOPY_SPA_CLIENT_SECRET'] = client_secret
            tenant_id = config['tenant']

            running_local = config['running_local']
            print(f"running_local:{running_local}")
            print(f"source_path:{source_path}")
            print(f"new_directory_name:{new_directory_name}")

            credential = ClientSecretCredential(tenant_id, client_id, client_secret)
            storage_account_loc = urlparse(source_path).netloc
            storage_path = urlparse(source_path).path
            storage_path_list = storage_path.split("/")
            storage_container = storage_path_list[1]
            account_url = f"https://{storage_account_loc}"

            service_client = DataLakeServiceClient(account_url=account_url,
                                                   credential=credential)
            file_system_client = service_client.get_file_system_client(storage_container)

            dir_path = storage_path.replace(f"{storage_container}" + "/", "")

            is_directory = None
            directory_client: DataLakeDirectoryClient
            try:
                directory_client = file_system_client.get_directory_client(dir_path)
                if directory_client.exists():
                    is_directory = True
                else:
                    is_directory = True

                if is_directory:
                    directory_client.rename_directory(new_directory_name)
                    result = "Success"
                else:
                    result = f"rename_directory failed: {dir_path} does not exist"
            except Exception as ex:
                directory_client = DataLakeDirectoryClient("empty", "empty", "empty")
                print(ex)
                result = "rename_directory failed"
        except Exception as ex_rename_directory:
            print(ex_rename_directory)
            result = "rename_directory failed"
        result = str(result)
        return result

    @classmethod
    def folder_adls_create(cls, config, dir_path: str, dbutils) -> str:

        running_local = config['running_local']
        client_id = config['client_id']
        client_secret = config['client_secret']

        if client_secret is None:
            azure_client_secret_key = str(config["azure_client_secret_key"])
            client_secret = f"Environment variable: {azure_client_secret_key} not found"

        os.environ['AZCOPY_SPA_CLIENT_SECRET'] = client_secret
        tenant_id = config['tenant']

        credential = ClientSecretCredential(tenant_id, client_id, client_secret)
        service_client = DataLakeServiceClient(account_url=account_url, credential=credential)
        file_system_client = service_client.get_file_system_client(storage_container)
        directory_client = file_system_client.create_directory(dir_path)

        return "True"

    @classmethod
    def file_adls_copy(cls, config, source_path: str, destination_path: str, from_to: str, dbutils) -> str:
        """copies local file to azure storage

        Args:
            config (dict): configuration dictionary
            source_path (str): source path to copy
            destination_path (str): destination path of copy or 'bytes' if should return byte array
            from_to (str): either BlobFSLocal or LocalBlobFS
            dbutils: Databricks dbutils object
        Returns:
            str: copy_status
        """
        result = "file_adls_copy failed"
        running_local = config['running_local']
        client_id = config['client_id']
        client_secret = config['client_secret']

        if client_secret is None:
            azure_client_secret_key = str(config["azure_client_secret_key"])
            client_secret = f"Environment variable: {azure_client_secret_key} not found"

        os.environ['AZCOPY_SPA_CLIENT_SECRET'] = client_secret
        tenant_id = config['tenant']

        print(f"running_local:{running_local}")
        print(f"from_to:{from_to}")
        print(f"source_path:{source_path}")
        print(f"destination_path:{destination_path}")

        if (running_local is True and (from_to == 'BlobFSLocal' or from_to == 'LocalBlobFS')):

            p_1 = f"--application-id={client_id}"
            p_2 = f"--tenant-id={tenant_id}"
            arr_azcopy_command = ["azcopy", "login", "--service-principal", p_1, p_2]
            arr_azcopy_command_string = ' '.join(arr_azcopy_command)
            print(arr_azcopy_command_string)

            try:
                check_output(arr_azcopy_command)
                result_1 = f"login --service-principal {p_1} to {p_2} succeeded"
            except subprocess.CalledProcessError as ex_called_process:
                result_1 = str(ex_called_process.output)

            print(result_1)

            if from_to == 'BlobFSLocal':
                arr_azcopy_command = [
                    'azcopy', 'copy', f"{source_path}", f"{destination_path}",
                    f'--from-to={from_to}', '--recursive',
                    '--trusted-microsoft-suffixes=', '--log-level=INFO']
            elif from_to == 'LocalBlobFS':
                arr_azcopy_command = [
                    'azcopy', 'copy', f"{source_path}", f"{destination_path}",
                    '--log-level=DEBUG', f'--from-to={from_to}']
            else:
                arr_azcopy_command = [f"from to:{from_to} is not supported"]

            arr_azcopy_command_string = ' '.join(arr_azcopy_command)
            print(arr_azcopy_command_string)

            try:
                check_output(arr_azcopy_command)
                result_2 = f"copy from {source_path} to {destination_path} succeeded"
            except subprocess.CalledProcessError as ex_called_process:
                result_2 = str(ex_called_process.output)

            result = result_1 + result_2
        elif ((running_local is False) and from_to == 'BlobFSLocal'):
            credential = ClientSecretCredential(tenant_id, client_id, client_secret)
            storage_account_loc = urlparse(source_path).netloc
            storage_path = urlparse(source_path).path
            storage_path_list = storage_path.split("/")
            storage_container = storage_path_list[1]
            account_url = f"https://{storage_account_loc}"
            service_client = DataLakeServiceClient(account_url=account_url, credential=credential)
            file_system_client = service_client.get_file_system_client(storage_container)
            dir_path = storage_path.replace(f"{storage_container}" + "/", "")
            is_directory = None
            directory_client: DataLakeDirectoryClient
            try:
                directory_client = file_system_client.get_directory_client(dir_path)
                if directory_client.exists():
                    is_directory = True
                else:
                    is_directory = True
            except Exception as ex:
                directory_client = DataLakeDirectoryClient("empty", "empty", "empty")
                print(ex)

            obj_repo = pade_repo.RepoCore()

            if is_directory is True:

                azure_files = []

                try:
                    azure_files = file_system_client.get_paths(path=dir_path)
                except Exception as ex:
                    print(ex)

                for file_path in azure_files:
                    print(str(f"file_path:{file_path}"))
                    file_path_name = file_path.name
                    file_name = os.path.basename(file_path_name)
                    file_client = directory_client.get_file_client(file_path)
                    file_data = file_client.download_file()
                    file_bytes = file_data.readall()
                    file_string = file_bytes.decode("utf-8")
                    first_200_chars_of_string = file_string[0:200]
                    destination_file_path = destination_path + "/" + file_path_name

                    if len(file_string) > 0:
                        try:
                            # os.remove(destination_file_path)
                            dbutils.fs.rm(destination_file_path)
                        except OSError as ex_os_error:
                            # if failed, report it back to the user
                            print(f"Error: {ex_os_error.filename} - {ex_os_error.strerror}.")
                        try:
                            print(f"dbutils.fs.put({destination_file_path}, {first_200_chars_of_string}, True)")
                            result = dbutils.fs.put(destination_file_path, file_string, True)
                        except Exception as ex_os_error:
                            # if failed, report it back to the user
                            print(f"Error: {ex_os_error}.")

                        content_type = "bytes"
                        result = obj_repo.import_file(config, file_bytes, content_type, destination_file_path)

                    else:
                        result = f"destination_file_path:{destination_file_path} is empty"
                    # file_to_copy = io.BytesIO(file_bytes)
                    # print(f"destination_file_path:{destination_file_path}")
            else:
                file_path = storage_path.replace(f"{storage_container}" + "/", "")
                print(f"file_path:{file_path}")
                file_client = file_system_client.get_file_client(file_path)
                file_data = file_client.download_file()
                file_bytes = file_data.readall()
                file_string = file_bytes.decode("utf-8")
                file_name = os.path.basename(file_path)
                destination_file_path = destination_path + "/" + file_name
                first_200_chars_of_string = file_string[0:500]
                if len(file_string) > 0:

                    try:
                        # os.remove(destination_file_path)
                        dbutils.fs.rm(destination_file_path)
                    except OSError as ex_os_error:
                        # if failed, report it back to the user
                        print(f"Error: {ex_os_error.filename}-{ex_os_error.strerror}.")

                    try:
                        print(f"dbutils.fs.put({destination_file_path}, {first_200_chars_of_string}, True)")
                        result = dbutils.fs.put(destination_file_path, file_string, True)
                    except Exception as ex_os_error:
                        # if failed, report it back to the user
                        print(f"Error: {ex_os_error}.")

                    content_type = "bytes"
                    result = obj_repo.import_file(config, file_bytes, content_type, destination_file_path)
                else:
                    result = f"destination_file_path:{destination_file_path} is empty"
        elif ((running_local is False) and from_to == 'LocalBlobFS'):
            url = destination_path
            storage_account_loc = urlparse(url).netloc
            storage_path = urlparse(url).path
            storage_path_list = storage_path.split("/")
            storage_container = storage_path_list[1]
            file_name = os.path.basename(destination_path)
            dir_path = storage_path.replace(file_name, "")
            dir_path = dir_path.replace(storage_container + '/', "")
            account_url = f"https://{storage_account_loc}"
            print(f"account_url:{account_url}")
            print(f"url:{url}")
            print(f"storage_path:{storage_path}")
            print(f"storage_container:{storage_container}")
            print(f"dir_path:{dir_path}")
            print(f"file_name:{file_name}")

            credential = ClientSecretCredential(tenant_id, client_id, client_secret)
            service_client = DataLakeServiceClient(account_url=account_url, credential=credential)
            file_system_client = service_client.get_file_system_client(storage_container)
            directory_client = file_system_client.get_directory_client(dir_path)
            file_client = directory_client.create_file(file_name)
            local_file = open(source_path, "r", encoding="utf-8")
            file_contents = local_file.read()
            file_client.append_data(data=file_contents, offset=0, length=len(file_contents))
            result = file_client.flush_data(len(file_contents))

            # with open(source_path) as f_json:
            #     json_data = json.load(f_json)
            # result = file_client.upload_data(json_data, overwrite=True, max_concurrency=5)

            # file_client = file_system_client.get_file_client(file_path)
            # file_data = file_client.download_file(0)
            # result = file_data.readall()

            # print(f" dbutils.fs.cp({source_path}, {destination_path})")
            # result = dbutils.fs.cp(source_path, destination_path)
        else:
            result = "invalid config: must download/client config files from azure to local"
            result = result + " - functionality not available on databricks"
            print(result)

        result = str(result)

        return result

    @classmethod
    def get_file_size(cls, running_local: bool, path: str, dbutils, spark: SparkSession) -> int:
        """Takes in file path, dbutils object and spark obejct, returns file size of provided path

        Args:
            path (str): path to check file size
            dbutils (object): Databricks dbutils object
            spark (SparkSession): spark session object

        Returns:
            int:  file size
        """

        if dbutils is not None:

            file_exists = cls.file_exists(running_local, path, dbutils)

            if file_exists is True:
                ddl_schema_3 = StructType(
                    [
                        StructField("path", StringType()),
                        StructField("name", StringType()),
                        StructField("size", IntegerType())
                    ]
                )

                ddl_schema_4 = StructType(
                    [
                        StructField("path", StringType()),
                        StructField("name", StringType()),
                        StructField("size", IntegerType()),
                        StructField("modification_time", LongType()),
                    ]
                )

                print(f"command: dbutils.fs.ls({path})")
                sk_list = dbutils.fs.ls(path)
                print(f"num_elements:{len(sk_list)}")

                df_file_list = None

                if len(sk_list) > 0:
                    if len(sk_list[0]) == 3:
                        df_file_list = spark.createDataFrame(sk_list, ddl_schema_3)
                    elif len(sk_list[0]) == 4:
                        df_file_list = spark.createDataFrame(sk_list, ddl_schema_4)

                    if df_file_list is None:
                        file_size = 0
                    else:
                        first = df_file_list.first()
                        if first is not None:
                            file_size = first.size
                        else:
                            file_size = -1

                    # df_file_list = df_file_list.toPandas()
                    # file_size = df_file_list.iloc[0, df_file_list.columns.get_loc("size")]

                    file_size = int(str(file_size))
                else:
                    file_size = -1
            else:
                file_size = -1
        else:
            file_size = -1

        return file_size

    @staticmethod
    def file_exists(running_local: bool, path: str, dbutils) -> bool:
        """Takes in path, dbutils object, returns whether file exists at provided path

        Args:
            running_local: bool
            path (str): path to file
            dbutils (object): databricks dbutils

        Returns:
            bool: True/False indication if file exists
        """

        if running_local is True:
            b_exists = os.path.exists(path)
        else:
            try:
                path = path.replace("/dbfs", "")
                if dbutils is not None:
                    dbutils.fs.ls(path)
                    b_exists = True
                else:
                    b_exists = False
            except Exception as exception_result:
                if "java.io.FileNotFoundException" in str(exception_result):
                    b_exists = False
                else:
                    b_exists = False
                    raise

        return b_exists

    @classmethod
    def copy_url_to_blob(cls, config: dict, src_url: str,
                         dest_path: str, file_name: str) -> str:
        """download file from source url and upload the file to the azure storage destination path

        Args:
            src_url (str): description
            dest_path (str): description
            file_name (str): description

        Returns:
            str: upload status message
        """

        info_message = f"copy_url_to_blob: src_url:{src_url}, dest_path:{dest_path}, file_name:{file_name}"
        print(info_message)

        client_id = config['client_id']
        client_secret = config['client_secret']
        tenant_id = config['tenant']
        if client_secret is None:
            azure_client_secret_key = str(config["azure_client_secret_key"])
            client_secret=f"Environment variable: {azure_client_secret_key}not found"
            print(client_secret)
        credential = ClientSecretCredential(tenant_id, client_id,
                                            client_secret)
        storage_account_loc = urlparse(dest_path).netloc
        storage_path = urlparse(dest_path).path
        storage_path_list = storage_path.split("/")
        storage_container = storage_path_list[1]
        account_url = f"https://{storage_account_loc}"
        service_client = DataLakeServiceClient(account_url=account_url, credential=credential)
        os.environ['AZCOPY_SPA_CLIENT_SECRET'] = client_secret
        dir_path = storage_path.replace(f"{storage_container}" + "/", "")
        print(f"dir_path:{dir_path}")
        file_system_client = service_client.get_file_system_client(storage_container)
        directory_client = file_system_client.get_directory_client(dir_path)
        file_response = requests.get(src_url)
        file_data = file_response.content
        try:
            file_client = directory_client.create_file(file_name)
            result = file_client.upload_data(file_data, overwrite=True, max_concurrency=5)
        except Exception as ex:
            print(ex)
            result = "upload failed"
        return result


